# -*- coding: utf-8 -*-

import io
import sys

import six

from brewtils.resolvers import ResolverBase


class ChunksResolver(ResolverBase):
    """Resolver that uses the Beergarden chunks API"""

    def __init__(self, easy_client):
        self.easy_client = easy_client

    def should_upload(self, value, definition):
        """
        Parameter type must be Base64 and the value must be either:

        - String representation of a valid filename.
        - An IOBase object
        """
        if definition.type.lower() == "base64":
            if sys.version_info[0] == 2:
                file_types = (io.IOBase, file)  # noqa: F821
            else:
                file_types = (io.IOBase,)

            if isinstance(value, six.string_types) or isinstance(value, file_types):
                return True

        return False

    def upload(self, value, definition):
        return self.easy_client.upload_chunked_file(value)

    def should_download(self, value, definition):
        return definition.type.lower() == "base64"

    def download(self, value, definition):
        return self.easy_client.download_chunked_file(value.id)



# -*- coding: utf-8 -*-
import json
import logging
import typing
from typing import Any, Optional, Union

import six
from box import Box

import brewtils.models
import brewtils.schemas
from brewtils.models import BaseModel

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable


class SchemaParser(object):
    """Serialize and deserialize Brewtils models"""

    _models = {
        "ChoicesSchema": brewtils.models.Choices,
        "CommandSchema": brewtils.models.Command,
        "CronTriggerSchema": brewtils.models.CronTrigger,
        "DateTriggerSchema": brewtils.models.DateTrigger,
        "EventSchema": brewtils.models.Event,
        "FileTriggerSchema": brewtils.models.FileTrigger,
        "GardenSchema": brewtils.models.Garden,
        "InstanceSchema": brewtils.models.Instance,
        "IntervalTriggerSchema": brewtils.models.IntervalTrigger,
        "JobSchema": brewtils.models.Job,
        "LoggingConfigSchema": brewtils.models.LoggingConfig,
        "QueueSchema": brewtils.models.Queue,
        "ParameterSchema": brewtils.models.Parameter,
        "PatchSchema": brewtils.models.PatchOperation,
        "PrincipalSchema": brewtils.models.Principal,
        "RefreshTokenSchema": brewtils.models.RefreshToken,
        "RequestSchema": brewtils.models.Request,
        "RequestFileSchema": brewtils.models.RequestFile,
        "FileSchema": brewtils.models.File,
        "FileChunkSchema": brewtils.models.FileChunk,
        "FileStatusSchema": brewtils.models.FileStatus,
        "RequestTemplateSchema": brewtils.models.RequestTemplate,
        "RoleSchema": brewtils.models.Role,
        "SystemSchema": brewtils.models.System,
        "OperationSchema": brewtils.models.Operation,
        "RunnerSchema": brewtils.models.Runner,
        "ResolvableSchema": brewtils.models.Resolvable,
    }

    logger = logging.getLogger(__name__)

    # Deserialization methods
    @classmethod
    def parse_system(cls, system, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a system model object

        Args:
            system: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A System object
        """
        return cls.parse(
            system, brewtils.models.System, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_instance(cls, instance, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to an instance model object

        Args:
            instance: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            An Instance object
        """
        return cls.parse(
            instance, brewtils.models.Instance, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_command(cls, command, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a command model object

        Args:
            command: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A Command object
        """
        return cls.parse(
            command, brewtils.models.Command, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_parameter(cls, parameter, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a parameter model object

        Args:
            parameter: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            An Parameter object
        """
        return cls.parse(
            parameter, brewtils.models.Parameter, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_request_file(cls, request_file, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a request file model object

        Args:
            request_file: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A RequestFile object
        """
        return cls.parse(
            request_file, brewtils.models.RequestFile, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_file(cls, file, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a  file model object

        Args:
            file: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A File object
        """
        return cls.parse(file, brewtils.models.File, from_string=from_string, **kwargs)

    @classmethod
    def parse_request(cls, request, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a request model object

        Args:
            request: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A Request object
        """
        return cls.parse(
            request, brewtils.models.Request, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_patch(cls, patch, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a patch model object

        .. note::
            for our patches, many is *always* set to True. We will always return a list
            from this method.

        Args:
            patch: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A PatchOperation object
        """
        return cls.parse(
            patch, brewtils.models.PatchOperation, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_logging_config(cls, logging_config, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a logging config model object

        Args:
            logging_config: The raw input
            from_string: True if 'input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A LoggingConfig object
        """
        return cls.parse(
            logging_config,
            brewtils.models.LoggingConfig,
            from_string=from_string,
            **kwargs
        )

    @classmethod
    def parse_event(cls, event, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to an event model object

        Args:
            event: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            An Event object
        """
        return cls.parse(
            event, brewtils.models.Event, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_queue(cls, queue, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a queue model object

        Args:
            queue: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A Queue object
        """
        return cls.parse(
            queue, brewtils.models.Queue, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_principal(cls, principal, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a principal model object

        Args:
            principal: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A Principal object
        """
        return cls.parse(
            principal, brewtils.models.Principal, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_role(cls, role, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a role model object

        Args:
            role: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A Role object
        """
        return cls.parse(role, brewtils.models.Role, from_string=from_string, **kwargs)

    @classmethod
    def parse_refresh_token(cls, refresh_token, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a refresh token object

        Args:
            refresh_token: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A RefreshToken object
        """
        return cls.parse(
            refresh_token,
            brewtils.models.RefreshToken,
            from_string=from_string,
            **kwargs
        )

    @classmethod
    def parse_job(cls, job, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a job model object

        Args:
            job: Raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A Job object.

        """
        return cls.parse(job, brewtils.models.Job, from_string=from_string, **kwargs)

    @classmethod
    def parse_garden(cls, garden, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a garden model object

        Args:
            garden: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A Garden object
        """
        return cls.parse(
            garden, brewtils.models.Garden, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_operation(cls, operation, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a garden model object

        Args:
            operation: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            An Operation object
        """
        return cls.parse(
            operation, brewtils.models.Operation, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_runner(cls, runner, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a runner model object

        Args:
            runner: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A Runner object
        """
        return cls.parse(
            runner, brewtils.models.Runner, from_string=from_string, **kwargs
        )

    @classmethod
    def parse_resolvable(cls, resolvable, from_string=False, **kwargs):
        """Convert raw JSON string or dictionary to a runner model object

        Args:
            resolvable: The raw input
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A Resolvable object
        """
        return cls.parse(
            resolvable, brewtils.models.Resolvable, from_string=from_string, **kwargs
        )

    @classmethod
    def parse(cls, data, model_class, from_string=False, **kwargs):
        """Convert a JSON string or dictionary into a model object

        Args:
            data: The raw input
            model_class: Class object of the desired model type
            from_string: True if input is a JSON string, False if a dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            A model object

        """
        if data is None:
            raise TypeError("Data can not be None")

        if from_string and not isinstance(data, six.string_types):
            raise TypeError("When from_string=True data must be a string-type")

        if model_class == brewtils.models.PatchOperation:
            if not kwargs.get("many", True):
                cls.logger.warning(
                    "A patch object should always be wrapped as a list of objects. "
                    "Thus, parsing will always return a list. You specified many as "
                    "False, this is being ignored and a list "
                    "will be returned anyway."
                )
            kwargs["many"] = True

        schema = getattr(brewtils.schemas, model_class.schema)(**kwargs)
        schema.context["models"] = cls._models

        return schema.loads(data).data if from_string else schema.load(data).data

    # Serialization methods
    @classmethod
    def serialize_system(cls, system, to_string=True, include_commands=True, **kwargs):
        """Convert a system model into serialized form

        Args:
            system: The system object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            include_commands: True if the system's command list should be included
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of system
        """
        if not include_commands:
            if "exclude" in kwargs:
                kwargs["exclude"] += ("commands",)
            else:
                kwargs["exclude"] = ("commands",)

        return cls.serialize(
            system,
            to_string=to_string,
            schema_name=brewtils.models.System.schema,
            **kwargs
        )

    @classmethod
    def serialize_instance(cls, instance, to_string=True, **kwargs):
        """Convert an instance model into serialized form

        Args:
            instance: The instance object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of instance
        """
        return cls.serialize(
            instance,
            to_string=to_string,
            schema_name=brewtils.models.Instance.schema,
            **kwargs
        )

    @classmethod
    def serialize_command(cls, command, to_string=True, **kwargs):
        """Convert a command model into serialized form

        Args:
            command: The command object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of command
        """
        return cls.serialize(
            command,
            to_string=to_string,
            schema_name=brewtils.models.Command.schema,
            **kwargs
        )

    @classmethod
    def serialize_parameter(cls, parameter, to_string=True, **kwargs):
        """Convert a parameter model into serialized form

        Args:
            parameter: The parameter object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of parameter
        """
        return cls.serialize(
            parameter,
            to_string=to_string,
            schema_name=brewtils.models.Parameter.schema,
            **kwargs
        )

    @classmethod
    def serialize_request_file(cls, request_file, to_string=True, **kwargs):
        """Convert a request file model into serialized form

        Args:
            request_file: The request file object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of request file
        """
        return cls.serialize(
            request_file,
            to_string=to_string,
            schema_name=brewtils.models.RequestFile.schema,
            **kwargs
        )

    @classmethod
    def serialize_request(cls, request, to_string=True, **kwargs):
        """Convert a request model into serialized form

        Args:
            request: The request object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of request
        """
        return cls.serialize(
            request,
            to_string=to_string,
            schema_name=brewtils.models.Request.schema,
            **kwargs
        )

    @classmethod
    def serialize_patch(cls, patch, to_string=True, **kwargs):
        """Convert a patch model into serialized form

        Args:
            patch: The patch object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of patch
        """
        return cls.serialize(
            patch,
            to_string=to_string,
            schema_name=brewtils.models.PatchOperation.schema,
            **kwargs
        )

    @classmethod
    def serialize_logging_config(cls, logging_config, to_string=True, **kwargs):
        """Convert a logging config model into serialize form

        Args:
            logging_config: The logging config object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of logging config
        """
        return cls.serialize(
            logging_config,
            to_string=to_string,
            schema_name=brewtils.models.LoggingConfig.schema,
            **kwargs
        )

    @classmethod
    def serialize_event(cls, event, to_string=True, **kwargs):
        """Convert a logging config model into serialized form

        Args:
            event: The event object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of event
        """
        return cls.serialize(
            event,
            to_string=to_string,
            schema_name=brewtils.models.Event.schema,
            **kwargs
        )

    @classmethod
    def serialize_queue(cls, queue, to_string=True, **kwargs):
        """Convert a queue model into serialized form

        Args:
            queue: The queue object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of queue
        """
        return cls.serialize(
            queue,
            to_string=to_string,
            schema_name=brewtils.models.Queue.schema,
            **kwargs
        )

    @classmethod
    def serialize_principal(cls, principal, to_string=True, **kwargs):
        """Convert a principal model into serialized form

        Args:
            principal: The principal object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation
        """
        return cls.serialize(
            principal,
            to_string=to_string,
            schema_name=brewtils.models.Principal.schema,
            **kwargs
        )

    @classmethod
    def serialize_role(cls, role, to_string=True, **kwargs):
        """Convert a role model into serialized form

        Args:
            role: The role object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation
        """
        return cls.serialize(
            role, to_string=to_string, schema_name=brewtils.models.Role.schema, **kwargs
        )

    @classmethod
    def serialize_refresh_token(cls, refresh_token, to_string=True, **kwargs):
        """Convert a role model into serialized form

        Args:
            refresh_token: The token object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation
        """
        return cls.serialize(
            refresh_token,
            to_string=to_string,
            schema_name=brewtils.models.RefreshToken.schema,
            **kwargs
        )

    @classmethod
    def serialize_job(cls, job, to_string=True, **kwargs):
        """Convert a job model into serialized form.

        Args:
            job: The job object(s) to be serialized.
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the shcema (e.g. many=True)

        Returns:
            Serialize representation of job.
        """
        return cls.serialize(
            job, to_string=to_string, schema_name=brewtils.models.Job.schema, **kwargs
        )

    @classmethod
    def serialize_garden(cls, garden, to_string=True, **kwargs):
        """Convert an garden model into serialized form

        Args:
            garden: The garden object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of garden
        """
        return cls.serialize(
            garden,
            to_string=to_string,
            schema_name=brewtils.models.Garden.schema,
            **kwargs
        )

    @classmethod
    def serialize_operation(cls, operation, to_string=True, **kwargs):
        """Convert an operation model into serialized form

        Args:
            operation: The operation object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of operation
        """
        return cls.serialize(
            operation,
            to_string=to_string,
            schema_name=brewtils.models.Operation.schema,
            **kwargs
        )

    @classmethod
    def serialize_runner(cls, runner, to_string=True, **kwargs):
        """Convert a runner model into serialized form

        Args:
            runner: The runner object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of runner
        """
        return cls.serialize(
            runner,
            to_string=to_string,
            schema_name=brewtils.models.Runner.schema,
            **kwargs
        )

    @classmethod
    def serialize_resolvable(cls, resolvable, to_string=True, **kwargs):
        """Convert a resolvable model into serialized form

        Args:
            resolvable: The resolvable object(s) to be serialized
            to_string: True to generate a JSON-formatted string, False to generate a
                dictionary
            **kwargs: Additional parameters to be passed to the Schema (e.g. many=True)

        Returns:
            Serialized representation of runner
        """
        return cls.serialize(
            resolvable,
            to_string=to_string,
            schema_name=brewtils.models.Resolvable.schema,
            **kwargs
        )

    @classmethod
    def serialize(
        cls,
        model,  # type: Union[BaseModel, typing.Iterable[BaseModel], dict]
        to_string=False,  # type: bool
        schema_name=None,  # type: str
        **kwargs  # type: Any
    ):
        # type: (...) -> Optional[str]
        """Convert a model object or list of models into a dictionary or JSON string.

        This is potentially recursive - here's how this should work:

        - Determine the correct schema to use for serializing. This can be explicitly
          passed as an argument, or it can be determined by inspecting the model to
          serialize.
        - Determine if the model to serialize is a collection or a single object.
            - If it's a single object, serialize it and return that.
            - If it's a collection, construct a list by calling this method for each
              individual item in the collection. Then serialize **that** and return it.

        Args:
            model: The model or model list
            to_string: True to generate a JSON string, False to generate a
                dictionary
            schema_name: Name of schema to use for serializing. If None, will be
            determined by inspecting ``model``
            **kwargs: Additional parameters to be passed to the Schema.
                Note that the 'many' parameter will be set correctly automatically.

        Returns:
            A serialized model representation

        """
        schema_name = schema_name or cls._get_schema_name(model)

        if cls._single_item(model):
            kwargs["many"] = False

            schema = getattr(brewtils.schemas, schema_name)(**kwargs)

            return schema.dumps(model).data if to_string else schema.dump(model).data

        # Explicitly force to_string to False so only original call returns a string
        multiple = [cls.serialize(x, to_string=False, **kwargs) for x in model]

        return json.dumps(multiple) if to_string else multiple

    @classmethod
    def _get_schema_name(cls, obj):
        # type: (Any) -> Optional[str]
        """Get the name of the schema to use for a particular object

        The model classes have a ``schema`` attribute with this info. We want to be able
        to pull this out for an instance of a model as well as the model class object.

        Args:
            obj: The object to inspect for a schema name

        Returns:
            The schema name, if found. None otherwise.
        """
        if isinstance(obj, brewtils.models.BaseModel):
            # Use type() here because Command has an instance attribute named "schema"
            return type(obj).schema

        return None

    @classmethod
    def _single_item(cls, obj):
        """Determine if the object given is a single item or a collection.

        For serialization to work correctly **all** these must work (on Python 2 & 3):

        - Brewtils models must return True
        - "Standard" collections (list, tuple, set) must return False
        - Dictionaries and Boxes must return True
        """
        if isinstance(obj, (dict, Box)):
            return True
        return not isinstance(obj, Iterable)



# -*- coding: utf-8 -*-
from brewtils.models import Resolvable
from brewtils.resolvers import ResolverBase


class IdentityResolver(ResolverBase):
    """Resolver that doesn't actually resolve anything

    On the upload side this is used to ensure that Resolvables always work when used in
    a SystemClient. For example, if you're using a SystemClient to execute a command
    with a Bytes parameter but you already have a Resolvable for that parameter, this
    makes that work.

    On the download side this is used to support autoresolve=False parameters. If a
    definition specifies "autoresolve": False as part of the type_info dictionary then
    the parameter WILL NOT be resolved before the command function is invoked. Instead,
    the Resolvable itself will be passed as that parameter. This might be useful if you
    wanted to farm out a bytes object to multiple children commands without needing to
    re-upload the same bytes every time.
    """

    def should_upload(self, value, definition):
        return isinstance(value, Resolvable)

    def upload(self, value, definition):
        return value

    def should_download(self, value, definition):
        return definition.type_info.get("autoresolve") is False

    def download(self, value, definition):
        return value



# -*- coding: utf-8 -*-

from enum import Enum

import pytz
import six
from brewtils.errors import ModelError, _deprecate

__all__ = [
    "BaseModel",
    "System",
    "Instance",
    "Command",
    "Parameter",
    "Request",
    "PatchOperation",
    "Choices",
    "LoggingConfig",
    "Event",
    "Events",
    "Queue",
    "Principal",
    "Role",
    "RefreshToken",
    "Job",
    "RequestFile",
    "File",
    "FileChunk",
    "FileStatus",
    "RequestTemplate",
    "DateTrigger",
    "CronTrigger",
    "IntervalTrigger",
    "FileTrigger",
    "Garden",
    "Operation",
    "Resolvable",
]


class Events(Enum):
    BREWVIEW_STARTED = 1
    BREWVIEW_STOPPED = 2
    BARTENDER_STARTED = 3
    BARTENDER_STOPPED = 4
    REQUEST_CREATED = 5
    REQUEST_STARTED = 6
    REQUEST_UPDATED = 22
    REQUEST_COMPLETED = 7
    REQUEST_CANCELED = 42
    INSTANCE_INITIALIZED = 8
    INSTANCE_STARTED = 9
    INSTANCE_UPDATED = 23
    INSTANCE_STOPPED = 10
    SYSTEM_CREATED = 11
    SYSTEM_UPDATED = 12
    SYSTEM_REMOVED = 13
    QUEUE_CLEARED = 14
    ALL_QUEUES_CLEARED = 15
    DB_CREATE = 16
    DB_UPDATE = 17
    DB_DELETE = 18
    GARDEN_CREATED = 19
    GARDEN_UPDATED = 20
    GARDEN_REMOVED = 21
    FILE_CREATED = 24
    GARDEN_STARTED = 25
    GARDEN_STOPPED = 26
    GARDEN_UNREACHABLE = 27
    GARDEN_ERROR = 28
    GARDEN_NOT_CONFIGURED = 29
    GARDEN_SYNC = 30
    ENTRY_STARTED = 31
    ENTRY_STOPPED = 32
    JOB_CREATED = 33
    JOB_DELETED = 34
    JOB_PAUSED = 35
    JOB_RESUMED = 36
    PLUGIN_LOGGER_FILE_CHANGE = 37
    RUNNER_STARTED = 38
    RUNNER_STOPPED = 39
    RUNNER_REMOVED = 40
    JOB_UPDATED = 41

    # Next: 43


class BaseModel(object):
    schema = None


class Command(BaseModel):
    schema = "CommandSchema"

    COMMAND_TYPES = ("ACTION", "INFO", "EPHEMERAL", "ADMIN")
    OUTPUT_TYPES = ("STRING", "JSON", "XML", "HTML", "JS", "CSS")

    def __init__(
        self,
        name=None,
        description=None,
        parameters=None,
        command_type=None,
        output_type=None,
        schema=None,
        form=None,
        template=None,
        icon_name=None,
        hidden=False,
        metadata=None,
    ):
        self.name = name
        self.description = description
        self.parameters = parameters or []
        self.command_type = command_type
        self.output_type = output_type
        self.schema = schema
        self.form = form
        self.template = template
        self.icon_name = icon_name
        self.hidden = hidden
        self.metadata = metadata or {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Command: %s>" % self.name

    def parameter_keys(self):
        """Get a list of all Parameter keys

        Returns:
            list[str]: A list containing each Parameter's key attribute
        """
        return [p.key for p in self.parameters]

    def parameter_keys_by_type(self, desired_type):
        """Get a list of all Parameter keys, filtered by Parameter type

        Args:
            desired_type (str): Parameter type

        Returns:
            list[str]: A list containing matching Parameters' key attribute
        """
        keys = []
        for param in self.parameters:
            key = param.keys_by_type(desired_type)
            if key:
                keys.append(key)
        return keys

    def get_parameter_by_key(self, key):
        """Lookup a Parameter using a given key

        Args:
            key (str): The Parameter key to use

        Returns:
            Parameter (Optional): A Parameter with the given key

            If a Parameter with the given key does not exist None will be returned.
        """
        for parameter in self.parameters:
            if parameter.key == key:
                return parameter

        return None

    def has_different_parameters(self, parameters):
        """Determine if parameters differ from the current parameters

        Args:
            parameters (Sequence[Parameter]): Parameter collection for comparison

        Returns:
            bool: True if the given Parameters differ, False if they are identical
        """
        if len(parameters) != len(self.parameters):
            return True

        for parameter in parameters:
            if parameter.key not in self.parameter_keys():
                return True

            current_param = self.get_parameter_by_key(parameter.key)
            if current_param.is_different(parameter):
                return True

        return False


class Instance(BaseModel):
    schema = "InstanceSchema"

    INSTANCE_STATUSES = {
        "INITIALIZING",
        "RUNNING",
        "PAUSED",
        "STOPPED",
        "DEAD",
        "UNRESPONSIVE",
        "STARTING",
        "STOPPING",
        "UNKNOWN",
    }

    def __init__(
        self,
        name=None,
        description=None,
        id=None,
        status=None,
        status_info=None,
        queue_type=None,
        queue_info=None,
        icon_name=None,
        metadata=None,
    ):
        self.name = name
        self.description = description
        self.id = id
        self.status = status.upper() if status else None
        self.status_info = status_info or {}
        self.queue_type = queue_type
        self.queue_info = queue_info or {}
        self.icon_name = icon_name
        self.metadata = metadata or {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Instance: name=%s, status=%s>" % (self.name, self.status)


class Choices(BaseModel):
    schema = "ChoicesSchema"

    TYPES = ("static", "url", "command")
    DISPLAYS = ("select", "typeahead")

    def __init__(self, type=None, display=None, value=None, strict=None, details=None):
        self.type = type
        self.strict = strict
        self.value = value
        self.display = display
        self.details = details or {}

    def __str__(self):
        return self.value.__str__()

    def __repr__(self):
        return "<Choices: type=%s, display=%s, value=%s>" % (
            self.type,
            self.display,
            self.value,
        )


class Parameter(BaseModel):
    schema = "ParameterSchema"

    TYPES = (
        "String",
        "Integer",
        "Float",
        "Boolean",
        "Any",
        "Dictionary",
        "Date",
        "DateTime",
        "Bytes",
        "Base64",
    )
    FORM_INPUT_TYPES = ("textarea",)

    def __init__(
        self,
        key=None,
        type=None,
        multi=None,
        display_name=None,
        optional=None,
        default=None,
        description=None,
        choices=None,
        parameters=None,
        nullable=None,
        maximum=None,
        minimum=None,
        regex=None,
        form_input_type=None,
        type_info=None,
        is_kwarg=None,
        model=None,
    ):
        self.key = key
        self.type = type
        self.multi = multi
        self.display_name = display_name
        self.optional = optional
        self.default = default
        self.description = description
        self.choices = choices
        self.parameters = parameters or []
        self.nullable = nullable
        self.maximum = maximum
        self.minimum = minimum
        self.regex = regex
        self.form_input_type = form_input_type
        self.type_info = type_info or {}

        # These are special - they aren't part of the Parameter "API" (they aren't in
        # the serialization schema) but we still need them on this model for consistency
        # when creating Clients - https://github.com/beer-garden/beer-garden/issues/777
        self.is_kwarg = is_kwarg
        self.model = model

    def __str__(self):
        return self.key

    def __repr__(self):
        return "<Parameter: key=%s, type=%s, description=%s>" % (
            self.key,
            self.type,
            self.description,
        )

    def keys_by_type(self, desired_type):
        """Gets all keys by the specified type.

        Since parameters can be nested, this method will also return all keys of all
        nested parameters. The return value is a possibly nested list, where the first
        value of each list is going to be a string, while the next value is a list.

        Args:
            desired_type (str): Desired type

        Returns:
            An empty list if the type does not exist, otherwise it will be a list
            containing at least one entry which is a string, each subsequent entry is a
            nested list with the same structure.
        """
        keys = []
        if self.type == desired_type:
            keys.append(self.key)

        if not self.parameters:
            return keys

        for param in self.parameters:
            nested_keys = param.keys_by_type(desired_type)
            if nested_keys:
                if not keys:
                    keys = [self.key]

                keys.append(nested_keys)
        return keys

    def is_different(self, other):
        if not type(other) is type(self):
            return True

        fields_to_compare = [
            "key",
            "type",
            "type_info",
            "multi",
            "optional",
            "default",
            "nullable",
            "maximum",
            "minimum",
            "regex",
        ]
        for field in fields_to_compare:
            if getattr(self, field) != getattr(other, field):
                return True

        if len(self.parameters) != len(other.parameters):
            return True

        parameter_keys = [p.key for p in self.parameters]
        for parameter in other.parameters:
            if parameter.key not in parameter_keys:
                return True

            current_param = list(
                filter((lambda p: p.key == parameter.key), self.parameters)
            )[0]
            if current_param.is_different(parameter):
                return True

        return False


class RequestFile(BaseModel):
    schema = "RequestFileSchema"

    def __init__(self, storage_type=None, filename=None, id=None):
        self.storage_type = storage_type
        self.filename = filename
        self.id = id

    def __str__(self):
        return self.filename

    def __repr__(self):
        return "<RequestFile: filename=%s, storage_type=%s>" % (
            self.filename,
            self.storage_type,
        )


class File(BaseModel):
    schema = "FileSchema"

    def __init__(
        self,
        id=None,
        owner_id=None,
        owner_type=None,
        updated_at=None,
        file_name=None,
        file_size=None,
        chunks=None,
        chunk_size=None,
        owner=None,
        job=None,
        request=None,
    ):
        self.id = id
        self.owner_id = owner_id
        self.owner_type = owner_type
        self.owner = owner
        self.job = job
        self.request = request
        self.updated_at = updated_at
        self.file_name = file_name
        self.file_size = file_size
        self.chunks = chunks
        self.chunk_size = chunk_size

    def __str__(self):
        return self.file_name

    def __repr__(self):
        return "<File: id=%s, file_name=%s, owner_id=%s>" % (
            self.id,
            self.file_name,
            self.owner_id,
        )


class FileChunk(BaseModel):
    schema = "FileChunkSchema"

    def __init__(self, id=None, file_id=None, offset=None, data=None, owner=None):
        self.id = id
        self.file_id = file_id
        self.offset = offset
        self.data = data
        self.owner = owner

    def __str__(self):
        return self.data

    def __repr__(self):
        return "<FileChunk: file_id=%s, offset=%s>" % (self.file_id, self.offset)


class FileStatus(BaseModel):
    schema = "FileStatusSchema"

    def __init__(
        self,
        owner_id=None,
        owner_type=None,
        updated_at=None,
        file_name=None,
        file_size=None,
        chunks=None,
        chunk_size=None,
        chunk_id=None,
        file_id=None,
        offset=None,
        data=None,
        valid=None,
        missing_chunks=None,
        expected_max_size=None,
        size_ok=None,
        expected_number_of_chunks=None,
        number_of_chunks=None,
        chunks_ok=None,
        operation_complete=None,
        message=None,
    ):
        # Top-level file info
        self.file_id = file_id
        self.file_name = file_name
        self.file_size = file_size
        self.updated_at = updated_at
        self.chunk_size = chunk_size
        self.chunks = chunks
        self.owner_id = owner_id
        self.owner_type = owner_type
        # Chunk info
        self.chunk_id = chunk_id
        self.offset = offset
        self.data = data
        # Validation metadata
        self.valid = valid
        self.missing_chunks = missing_chunks
        self.expected_number_of_chunks = expected_number_of_chunks
        self.expected_max_size = expected_max_size
        self.number_of_chunks = number_of_chunks
        self.size_ok = size_ok
        self.chunks_ok = chunks_ok
        self.operation_complete = operation_complete
        self.message = message

    def __str__(self):
        return "%s" % self.__dict__

    def __repr__(self):
        return "<FileStatus: %s>" % self.__dict__


class RequestTemplate(BaseModel):
    schema = "RequestTemplateSchema"

    TEMPLATE_FIELDS = [
        "system",
        "system_version",
        "instance_name",
        "namespace",
        "command",
        "command_type",
        "parameters",
        "comment",
        "metadata",
        "output_type",
    ]

    def __init__(
        self,
        system=None,
        system_version=None,
        instance_name=None,
        namespace=None,
        command=None,
        command_type=None,
        parameters=None,
        comment=None,
        metadata=None,
        output_type=None,
    ):
        self.system = system
        self.system_version = system_version
        self.instance_name = instance_name
        self.namespace = namespace
        self.command = command
        self.command_type = command_type
        self.parameters = parameters
        self.comment = comment
        self.metadata = metadata or {}
        self.output_type = output_type

    def __str__(self):
        return self.command

    def __repr__(self):
        return (
            "<RequestTemplate: command=%s, system=%s, system_version=%s, "
            "instance_name=%s, namespace=%s>"
            % (
                self.command,
                self.system,
                self.system_version,
                self.instance_name,
                self.namespace,
            )
        )


class Request(RequestTemplate):
    schema = "RequestSchema"

    STATUS_LIST = (
        "CREATED",
        "RECEIVED",
        "IN_PROGRESS",
        "CANCELED",
        "SUCCESS",
        "ERROR",
        "INVALID",
    )
    COMPLETED_STATUSES = ("CANCELED", "SUCCESS", "ERROR", "INVALID")
    COMMAND_TYPES = ("ACTION", "INFO", "EPHEMERAL", "ADMIN")
    OUTPUT_TYPES = ("STRING", "JSON", "XML", "HTML", "JS", "CSS")

    def __init__(
        self,
        system=None,
        system_version=None,
        instance_name=None,
        namespace=None,
        command=None,
        id=None,
        parent=None,
        children=None,
        parameters=None,
        comment=None,
        output=None,
        output_type=None,
        status=None,
        command_type=None,
        created_at=None,
        error_class=None,
        metadata=None,
        hidden=None,
        updated_at=None,
        has_parent=None,
        requester=None,
    ):
        super(Request, self).__init__(
            system=system,
            system_version=system_version,
            instance_name=instance_name,
            namespace=namespace,
            command=command,
            command_type=command_type,
            parameters=parameters,
            comment=comment,
            metadata=metadata,
            output_type=output_type,
        )
        self.id = id
        self.parent = parent
        self.children = children
        self.output = output
        self._status = status
        self.hidden = hidden
        self.created_at = created_at
        self.updated_at = updated_at
        self.error_class = error_class
        self.has_parent = has_parent
        self.requester = requester

    @classmethod
    def from_template(cls, template, **kwargs):
        """Create a Request instance from a RequestTemplate

        Args:
            template: The RequestTemplate to use
            **kwargs: Optional overrides to use in place of the template's attributes

        Returns:
            The new Request instance
        """
        request_params = {
            k: kwargs.get(k, getattr(template, k))
            for k in RequestTemplate.TEMPLATE_FIELDS
        }
        return Request(**request_params)

    def __repr__(self):
        return (
            "<Request: command=%s, status=%s, system=%s, system_version=%s, "
            "instance_name=%s, namespace=%s>"
            % (
                self.command,
                self.status,
                self.system,
                self.system_version,
                self.instance_name,
                self.namespace,
            )
        )

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

    @property
    def is_ephemeral(self):
        return self.command_type and self.command_type.upper() == "EPHEMERAL"

    @property
    def is_json(self):
        return self.output_type and self.output_type.upper() == "JSON"


class System(BaseModel):
    schema = "SystemSchema"

    def __init__(
        self,
        name=None,
        description=None,
        version=None,
        id=None,
        max_instances=None,
        instances=None,
        commands=None,
        icon_name=None,
        display_name=None,
        metadata=None,
        namespace=None,
        local=None,
        template=None,
    ):
        self.name = name
        self.description = description
        self.version = version
        self.id = id
        self.max_instances = max_instances
        self.instances = instances or []
        self.commands = commands or []
        self.icon_name = icon_name
        self.display_name = display_name
        self.metadata = metadata or {}
        self.namespace = namespace
        self.local = local
        self.template = template

    def __str__(self):
        return "%s:%s-%s" % (self.namespace, self.name, self.version)

    def __repr__(self):
        return "<System: name=%s, version=%s, namespace=%s>" % (
            self.name,
            self.version,
            self.namespace,
        )

    @property
    def instance_names(self):
        return [i.name for i in self.instances]

    def has_instance(self, name):
        """Determine if an instance currently exists in the system

        Args:
            name (str): The instance name

        Returns:
            bool: True if an instance with the given name exists, False otherwise
        """
        return name in self.instance_names

    def get_instance_by_name(self, name, raise_missing=False):
        """Get an instance that currently exists in the system

        Args:
            name (str): The instance name
            raise_missing (bool): If True, raise an exception if an Instance with the
            given name is not found. If False, will return None in that case.

        Returns:
            Instance: The instance if it exists, None otherwise

        Raises:
            ModelError: Instance was not found and raise_missing=True
        """
        for instance in self.instances:
            if instance.name == name:
                return instance

        if raise_missing:
            raise ModelError("Instance not found")

        return None

    def get_instance_by_id(self, id, raise_missing=False):
        """Get an instance that currently exists in the system

        Args:
            id (str): The instance id
            raise_missing (bool): If True, raise an exception if an Instance with the
            given id is not found. If False, will return None in that case.

        Returns:
            Instance: The instance if it exists, None otherwise

        Raises:
            ModelError: Instance was not found and raise_missing=True
        """
        for instance in self.instances:
            if instance.id == id:
                return instance

        if raise_missing:
            raise ModelError("Instance not found")

        return None

    def get_instance(self, name):
        """
        .. deprecated::3.0
           Will be removed in 4.0. Use ``get_instance_by_name`` instead
        """
        _deprecate(
            "Heads up! This method is deprecated, please use get_instance_by_name"
        )
        return self.get_instance_by_name(name)

    def get_command_by_name(self, command_name):
        """Retrieve a particular command from the system

        Args:
            command_name (str): The command name

        Returns:
            Command: The command if it exists, None otherwise
        """
        for command in self.commands:
            if command.name == command_name:
                return command

        return None

    def has_different_commands(self, commands):
        """Check if a set of commands is different than the current commands

        Args:
            commands (Sequence[Command]): Command collection for comparison

        Returns:
            bool: True if the given Commands differ, False if they are identical
        """
        if len(commands) != len(self.commands):
            return True

        for command in commands:
            if command.name not in [c.name for c in self.commands]:
                return True

            current_command = self.get_command_by_name(command.name)

            if current_command.has_different_parameters(command.parameters):
                return True

        return False


class PatchOperation(BaseModel):
    schema = "PatchSchema"

    def __init__(self, operation=None, path=None, value=None):
        self.operation = operation
        self.path = path
        self.value = value

    def __str__(self):
        return "%s, %s, %s" % (self.operation, self.path, self.value)

    def __repr__(self):
        return "<Patch: operation=%s, path=%s, value=%s>" % (
            self.operation,
            self.path,
            self.value,
        )


class LoggingConfig(BaseModel):
    schema = "LoggingConfigSchema"

    LEVELS = ("DEBUG", "INFO", "WARN", "ERROR")
    SUPPORTED_HANDLERS = ("stdout", "file", "logstash")

    DEFAULT_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    DEFAULT_HANDLER = {
        "class": "logging.StreamHandler",
        "stream": "ext::/sys.stdout",
        "formatter": "default",
    }

    def __init__(self, level=None, handlers=None, formatters=None, loggers=None):
        self.level = level
        self.handlers = handlers
        self.formatters = formatters
        self._loggers = loggers or {}

    @property
    def handler_names(self):
        if self.handlers:
            return set(self.handlers)
        else:
            return None

    @property
    def formatter_names(self):
        if self.formatters:
            return set(self.formatters)
        else:
            return None

    def get_plugin_log_config(self, **kwargs):
        """Get a specific plugin logging configuration.

        It is possible for different systems to have different logging configurations.
        This method will create the correct plugin logging configuration and return it.
        If a specific logger is not found for a system, then the current logging
        configuration will be returned.

        Keyword Args:
            Identifying information for a system (i.e. system_name)

        Returns:
            The logging configuration for this system
        """
        system_name = kwargs.pop("system_name", None)
        specific_logger = self._loggers.get(system_name, {})

        # If there is no specific logger, then we simply return this object
        # otherwise, we need to construct a new LoggingConfig object with
        # the overrides given in the logger.
        if not specific_logger:
            return self

        level = specific_logger.get("level", self.level)
        handlers = self._generate_handlers(specific_logger.get("handlers"))
        formatters = self._generate_formatters(specific_logger.get("formatters", {}))

        return LoggingConfig(level=level, handlers=handlers, formatters=formatters)

    def _generate_handlers(self, specific_handlers):

        # If we are not given an override for handlers, then we will just
        # assume that we want to use all the handlers given in the current
        # configuration.
        if not specific_handlers:
            return self.handlers

        if isinstance(specific_handlers, list):
            handlers = {}
            for handler_name in specific_handlers:
                handlers[handler_name] = self.handlers[handler_name]
        else:
            return specific_handlers

        return handlers

    def _generate_formatters(self, specific_formatters):

        # If we are not given an override for formatters, then we will just
        # assume that we want to use the formatters given in the current
        # configuration
        if not specific_formatters:
            return self.formatters

        # In case no formatter is provided, we always want a default.
        formatters = {"default": {"format": self.DEFAULT_FORMAT}}
        for formatter_name, format_str in six.iteritems(specific_formatters):
            formatters[formatter_name] = {"format": format_str}

        return formatters

    def __str__(self):
        return "%s, %s, %s" % (self.level, self.handler_names, self.formatter_names)

    def __repr__(self):
        return "<LoggingConfig: level=%s, handlers=%s, formatters=%s" % (
            self.level,
            self.handler_names,
            self.formatter_names,
        )


class Event(BaseModel):
    schema = "EventSchema"

    def __init__(
        self,
        name=None,
        namespace=None,
        garden=None,
        metadata=None,
        timestamp=None,
        payload_type=None,
        payload=None,
        error=None,
        error_message=None,
    ):
        self.name = name
        self.namespace = namespace
        self.garden = garden
        self.metadata = metadata or {}
        self.timestamp = timestamp
        self.payload_type = payload_type
        self.payload = payload
        self.error = error
        self.error_message = error_message

    def __str__(self):
        return "%s: %s" % (self.namespace, self.name)

    def __repr__(self):
        return (
            "<Event: namespace=%s, garden=%s, name=%s, timestamp=%s, error=%s, "
            "error_message=%s, metadata=%s, payload_type=%s, payload=%r>"
            % (
                self.namespace,
                self.garden,
                self.name,
                self.timestamp,
                self.error,
                self.error_message,
                self.metadata,
                self.payload_type,
                self.payload,
            )
        )


class Queue(BaseModel):
    schema = "QueueSchema"

    def __init__(
        self,
        name=None,
        system=None,
        version=None,
        instance=None,
        system_id=None,
        display=None,
        size=None,
    ):
        self.name = name
        self.system = system
        self.version = version
        self.instance = instance
        self.system_id = system_id
        self.display = display
        self.size = size

    def __str__(self):
        return "%s: %s" % (self.name, self.size)

    def __repr__(self):
        return "<Queue: name=%s, size=%s>" % (self.name, self.size)


class Principal(BaseModel):
    schema = "PrincipalSchema"

    def __init__(
        self,
        id=None,
        username=None,
        roles=None,
        permissions=None,
        preferences=None,
        metadata=None,
    ):
        self.id = id
        self.username = username
        self.roles = roles
        self.permissions = permissions
        self.preferences = preferences
        self.metadata = metadata

    def __str__(self):
        return "%s" % self.username

    def __repr__(self):
        return "<Principal: username=%s, roles=%s, permissions=%s>" % (
            self.username,
            self.roles,
            self.permissions,
        )


class Role(BaseModel):
    schema = "RoleSchema"

    def __init__(self, id=None, name=None, description=None, permissions=None):
        self.id = id
        self.name = name
        self.description = description
        self.permissions = permissions

    def __str__(self):
        return "%s" % self.name

    def __repr__(self):
        return "<Role: name=%s, permissions=%s>" % (self.name, self.permissions)


class RefreshToken(BaseModel):
    schema = "RefreshTokenSchema"

    def __init__(self, id=None, issued=None, expires=None, payload=None):
        self.id = id
        self.issued = issued
        self.expires = expires
        self.payload = payload or {}

    def __str__(self):
        return "%s" % self.payload

    def __repr__(self):
        return "<RefreshToken: issued=%s, expires=%s, payload=%s>" % (
            self.issued,
            self.expires,
            self.payload,
        )


class Job(BaseModel):
    TRIGGER_TYPES = {"interval", "date", "cron", "file"}
    STATUS_TYPES = {"RUNNING", "PAUSED"}
    schema = "JobSchema"

    def __init__(
        self,
        id=None,
        name=None,
        trigger_type=None,
        trigger=None,
        request_template=None,
        misfire_grace_time=None,
        coalesce=None,
        next_run_time=None,
        success_count=None,
        error_count=None,
        status=None,
        max_instances=None,
        timeout=None,
    ):
        self.id = id
        self.name = name
        self.trigger_type = trigger_type
        self.trigger = trigger
        self.request_template = request_template
        self.misfire_grace_time = misfire_grace_time
        self.coalesce = coalesce
        self.next_run_time = next_run_time
        self.success_count = success_count
        self.error_count = error_count
        self.status = status
        self.max_instances = max_instances
        self.timeout = timeout

    def __str__(self):
        return "%s: %s" % (self.name, self.id)

    def __repr__(self):
        return "<Job: name=%s, id=%s>" % (self.name, self.id)


class DateTrigger(BaseModel):
    schema = "DateTriggerSchema"

    def __init__(self, run_date=None, timezone=None):
        self.run_date = run_date
        self.timezone = timezone

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "<DateTrigger: run_date=%s>" % self.run_date

    @property
    def scheduler_attributes(self):
        return ["run_date", "timezone"]

    @property
    def scheduler_kwargs(self):
        tz = pytz.timezone(self.timezone)

        return {"timezone": tz, "run_date": tz.localize(self.run_date)}


class IntervalTrigger(BaseModel):
    schema = "IntervalTriggerSchema"

    def __init__(
        self,
        weeks=None,
        days=None,
        hours=None,
        minutes=None,
        seconds=None,
        start_date=None,
        end_date=None,
        timezone=None,
        jitter=None,
        reschedule_on_finish=None,
    ):
        self.weeks = weeks
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds
        self.start_date = start_date
        self.end_date = end_date
        self.timezone = timezone
        self.jitter = jitter
        self.reschedule_on_finish = reschedule_on_finish

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            "<IntervalTrigger: weeks=%d, days=%d, hours=%d, "
            "minutes=%d, seconds=%d>"
            % (self.weeks, self.days, self.hours, self.minutes, self.seconds)
        )

    @property
    def scheduler_attributes(self):
        return [
            "weeks",
            "days",
            "hours",
            "minutes",
            "seconds",
            "start_date",
            "end_date",
            "timezone",
            "jitter",
            "reschedule_on_finish",
        ]

    @property
    def scheduler_kwargs(self):
        tz = pytz.timezone(self.timezone)

        kwargs = {key: getattr(self, key) for key in self.scheduler_attributes}
        kwargs.update(
            {
                "timezone": tz,
                "start_date": tz.localize(self.start_date) if self.start_date else None,
                "end_date": tz.localize(self.end_date) if self.end_date else None,
            }
        )

        return kwargs


class CronTrigger(BaseModel):
    schema = "CronTriggerSchema"

    def __init__(
        self,
        year=None,
        month=None,
        day=None,
        week=None,
        day_of_week=None,
        hour=None,
        minute=None,
        second=None,
        start_date=None,
        end_date=None,
        timezone=None,
        jitter=None,
    ):
        self.year = year
        self.month = month
        self.day = day
        self.week = week
        self.day_of_week = day_of_week
        self.hour = hour
        self.minute = minute
        self.second = second
        self.start_date = start_date
        self.end_date = end_date
        self.timezone = timezone
        self.jitter = jitter

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "<CronTrigger: %s %s %s %s %s>" % (
            self.minute,
            self.hour,
            self.day,
            self.month,
            self.day,
        )

    @property
    def scheduler_attributes(self):
        return [
            "year",
            "month",
            "day",
            "week",
            "day_of_week",
            "hour",
            "minute",
            "second",
            "start_date",
            "end_date",
            "timezone",
            "jitter",
        ]

    @property
    def scheduler_kwargs(self):
        tz = pytz.timezone(self.timezone)

        kwargs = {key: getattr(self, key) for key in self.scheduler_attributes}
        kwargs.update(
            {
                "timezone": tz,
                "start_date": tz.localize(self.start_date) if self.start_date else None,
                "end_date": tz.localize(self.end_date) if self.end_date else None,
            }
        )

        return kwargs


class FileTrigger(BaseModel):
    schema = "FileTriggerSchema"

    def __init__(self, pattern=None, path=None, recursive=None, callbacks=None):
        self.pattern = pattern
        self.path = path
        self.recursive = recursive
        self.callbacks = callbacks

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "<FileTrigger: %s %s %s %s>" % (
            self.pattern,
            self.path,
            self.recursive,
            self.callbacks,
        )

    @property
    def scheduler_attributes(self):
        return ["pattern", "path", "recursive", "callbacks"]

    @property
    def scheduler_kwargs(self):
        kwargs = {key: getattr(self, key) for key in self.scheduler_attributes}
        kwargs.update(
            {
                "pattern": self.pattern,
                "path": self.path,
                "recursive": self.recursive,
                "callbacks": self.callbacks,
            }
        )

        return kwargs


class Garden(BaseModel):
    schema = "GardenSchema"

    GARDEN_STATUSES = {
        "INITIALIZING",
        "RUNNING",
        "BLOCKED",
        "STOPPED",
        "NOT_CONFIGURED",
        "UNREACHABLE",
        "ERROR",
        "UNKNOWN",
    }

    def __init__(
        self,
        id=None,
        name=None,
        status=None,
        status_info=None,
        namespaces=None,
        systems=None,
        connection_type=None,
        connection_params=None,
    ):
        self.id = id
        self.name = name
        self.status = status.upper() if status else None
        self.status_info = status_info or {}
        self.namespaces = namespaces or []
        self.systems = systems or []

        self.connection_type = connection_type
        self.connection_params = connection_params

    def __str__(self):
        return "%s" % self.name

    def __repr__(self):
        return "<Garden: garden_name=%s, status=%s>" % (self.name, self.status)


class Operation(BaseModel):
    schema = "OperationSchema"

    def __init__(
        self,
        model=None,
        model_type=None,
        args=None,
        kwargs=None,
        target_garden_name=None,
        source_garden_name=None,
        operation_type=None,
    ):
        self.model = model
        self.model_type = model_type
        self.args = args or []
        self.kwargs = kwargs or {}
        self.target_garden_name = target_garden_name
        self.source_garden_name = source_garden_name
        self.operation_type = operation_type

    def __str__(self):
        return "%s" % self.operation_type

    def __repr__(self):
        return (
            "<Operation: operation_type=%s, source_garden_name=%s, "
            "target_garden_name=%s, model_type=%s, model=%s, args=%s, kwargs=%s>"
            % (
                self.operation_type,
                self.source_garden_name,
                self.target_garden_name,
                self.model_type,
                self.model,
                self.args,
                self.kwargs,
            )
        )


class Runner(BaseModel):
    schema = "RunnerSchema"

    def __init__(
        self,
        id=None,
        name=None,
        path=None,
        instance_id=None,
        stopped=None,
        dead=None,
        restart=None,
    ):
        self.id = id
        self.name = name
        self.path = path
        self.instance_id = instance_id
        self.stopped = stopped
        self.dead = dead
        self.restart = restart

    def __str__(self):
        return "%s" % self.name

    def __repr__(self):
        return (
            "<Runner: id=%s, name=%s, path=%s, instance_id=%s, stopped=%s, dead=%s, "
            "restart=%s>"
            % (
                self.id,
                self.name,
                self.path,
                self.instance_id,
                self.stopped,
                self.dead,
                self.restart,
            )
        )


class Resolvable(BaseModel):
    schema = "ResolvableSchema"

    # Resolvable parameter types
    TYPES = ("Base64", "Bytes")

    def __init__(self, id=None, type=None, storage=None, details=None):
        self.id = id
        self.type = type
        self.storage = storage
        self.details = details or {}

    def __str__(self):
        return "%s: %s %s" % (self.id, self.type, self.storage)

    def __repr__(self):
        return "<Resolvable: id=%s, type=%s, storage=%s, details=%s>" % (
            self.id,
            self.type,
            self.storage,
            self.details,
        )



# -*- coding: utf-8 -*-

from brewtils.resolvers import ResolverBase


class BytesResolver(ResolverBase):
    """Resolver that uses the Beergarden file API"""

    def __init__(self, easy_client):
        self.easy_client = easy_client

    def should_upload(self, value, definition):
        return definition.type.lower() == "bytes"

    def upload(self, value, definition):
        return self.easy_client.upload_bytes(value)

    def should_download(self, value, definition):
        return definition.type.lower() == "bytes"

    def download(self, value, definition):
        return self.easy_client.download_bytes(value.id)



# -*- coding: utf-8 -*-

import json
import os
from io import open
from typing import Optional, Union

import six

from brewtils.errors import PluginParamError


def resolve_schema(schema=None, base_dir=None):
    # type: (Union[dict, str], str) -> Optional[dict]
    """Resolve a schema attribute

    Returns:
        Dictionary that fully describes a schema
    """
    if schema is None or isinstance(schema, dict):
        return schema
    elif isinstance(schema, six.string_types):
        try:
            if schema.startswith("http"):
                return _load_from_url(schema)
            elif schema.startswith("/") or schema.startswith("."):
                return json.loads(_load_from_path(schema, base_dir=base_dir))
            else:
                raise PluginParamError("Schema was not a definition, file path, or URL")
        except Exception as ex:
            six.raise_from(PluginParamError("Error resolving schema: %s" % ex), ex)
    else:
        raise PluginParamError(
            "Schema specified was not a definition, file path, or URL"
        )


def resolve_form(form=None, base_dir=None):
    # type: (Union[None, dict, list, str], str) -> Optional[dict]
    """Resolve a form attribute

    Returns:
        Dictionary that fully describes a form
    """
    if form is None or isinstance(form, dict):
        return form
    elif isinstance(form, list):
        return {"type": "fieldset", "items": form}
    elif isinstance(form, six.string_types):
        try:
            if form.startswith("http"):
                return _load_from_url(form)
            elif form.startswith("/") or form.startswith("."):
                return json.loads(_load_from_path(form, base_dir=base_dir))
            else:
                raise PluginParamError("Form was not a definition, file path, or URL")
        except Exception as ex:
            six.raise_from(PluginParamError("Error resolving form: %s" % ex), ex)
    else:
        raise PluginParamError("Schema was not a definition, file path, or URL")


def resolve_template(template=None, base_dir=None):
    # type: (str, str) -> Optional[str]
    """Resolve a template attribute

    Returns:
        Dictionary that fully describes a template
    """
    if template is None:
        return None
    elif isinstance(template, six.string_types):
        try:
            if template.startswith("http"):
                return _load_from_url(template)
            elif template.startswith("/") or template.startswith("."):
                return _load_from_path(template, base_dir=base_dir)
            else:
                return template

        except Exception as ex:
            six.raise_from(PluginParamError("Error resolving template: %s" % ex), ex)
    else:
        raise PluginParamError(
            "Template specified was not a definition, file path, or URL"
        )


def _load_from_url(url):
    # type: (str) -> Union[str, dict]
    """Load a definition from a URL"""
    from brewtils.rest.client import RestClient

    # Use a RestClient's session since TLS will (hopefully) be configured
    response = RestClient(bg_host="").session.get(url)

    if "application/json" in response.headers.get("content-type", "").lower():
        return response.json()
    return response.text


def _load_from_path(path, base_dir=None):
    # type: (str, str) -> str
    """Load a definition from a path

    This is a little odd because of the differences between command-level resources and
    system-level ones and the need to be backwards-compatible.

    The problem is determining the "correct" base to use when a relative path is given.
    For command resources we're able to use ``inspect`` on the method object to
    determine the current module file. So we've always just used the directory
    containing that file as the base for relative paths.

    However, we don't have that ability when it comes to system resources. For example,
    supposed the system template is specified in an environment variable:

      BG_SYSTEM_TEMPLATE=./resources/template.html

    The only thing that really makes sense in this case is starting in the plugin's
    current working directory. In hindsight, we probably should have done that for
    command resources as well.

    Finally, having ONLY system resources start at cwd and ONLY command resources be
    dependent on the file in which they're declared would be extremely confusing.

    So in an attempt to remain compatible this will attempt to use a provided base_dir
    as the starting point for relative path resolution. If there's no file there then it
    will re-resolve the path, this time using the cwd as he starting point. If no
    base_dir is provided then it'll just use the cwd.

    Going forward this will hopefully allow us to present using the cwd as a best
    practice for everything without forcing anyone to rewrite their plugins.
    """

    if base_dir and os.path.exists(os.path.abspath(os.path.join(base_dir, path))):
        file_path = os.path.abspath(os.path.join(base_dir, path))
    else:
        file_path = os.path.abspath(os.path.join(os.getcwd(), path))

    try:
        with open(file_path, "r") as definition_file:
            return definition_file.read()
    except IOError as ex:
        six.raise_from(
            PluginParamError(
                "%s. Please remember that relative paths will be resolved starting "
                "from the plugin's current working directory." % ex
            ),
            ex,
        )



# -*- coding: utf-8 -*-
import json
import logging
import logging.config
import os
import signal
import sys
import threading

import appdirs
from box import Box
from packaging.version import Version
from pathlib import Path
from requests import ConnectionError as RequestsConnectionError

import brewtils
from brewtils.config import load_config
from brewtils.decorators import _parse_client
from brewtils.display import resolve_template
from brewtils.errors import (
    ConflictError,
    DiscardMessageException,
    PluginValidationError,
    RequestProcessingError,
    RestConnectionError,
    ValidationError,
    _deprecate,
)
from brewtils.log import configure_logging, default_config, find_log_file, read_log_file
from brewtils.models import Instance, System
from brewtils.request_handling import (
    AdminProcessor,
    HTTPRequestUpdater,
    RequestConsumer,
    RequestProcessor,
)
from brewtils.resolvers.manager import ResolutionManager
from brewtils.rest.easy_client import EasyClient
from brewtils.specification import _CONNECTION_SPEC

# This is what enables request nesting to work easily
request_context = threading.local()
request_context.current_request = None

# Global config, used to simplify BG client creation and sanity checks.
CONFIG = Box(default_box=True)


class Plugin(object):
    """A Beer-garden Plugin

    This class represents a Beer-garden Plugin - a continuously-running process
    that can receive and process Requests.

    To work, a Plugin needs a Client instance - an instance of a class defining
    which Requests this plugin can accept and process. The easiest way to define
    a ``Client`` is by annotating a class with the ``@system`` decorator.

    A Plugin needs certain pieces of information in order to function correctly. These
    can be grouped into two high-level categories: identifying information and
    connection information.

    Identifying information is how Beer-garden differentiates this Plugin from all
    other Plugins. If you already have fully-defined System model you can pass that
    directly to the Plugin (``system=my_system``). However, normally it's simpler to
    pass the pieces directly:

        - ``name`` (required)
        - ``version`` (required)
        - ``instance_name`` (required, but defaults to "default")
        - ``namespace``
        - ``description``
        - ``icon_name``
        - ``metadata``
        - ``display_name``

    Connection information tells the Plugin how to communicate with Beer-garden. The
    most important of these is the ``bg_host`` (to tell the plugin where to find the
    Beer-garden you want to connect to):

        - ``bg_host``
        - ``bg_port``
        - ``bg_url_prefix``
        - ``ssl_enabled``
        - ``ca_cert``
        - ``ca_verify``
        - ``client_cert``

    An example plugin might look like this:

    .. code-block:: python

        Plugin(
            name="Test",
            version="1.0.0",
            instance_name="default",
            namespace="test plugins",
            description="A Test",
            bg_host="localhost",
        )

    Plugins use `Yapconf <https://github.com/loganasherjones/yapconf>`_ for
    configuration loading, which means that values can be discovered from sources other
    than direct argument passing. Config can be passed as command line arguments::

        python my_plugin.py --bg-host localhost

    Values can also be specified as environment variables with a "\\BG_" prefix::

        BG_HOST=localhost python my_plugin.py

    Plugins service requests using a
    :py:class:`concurrent.futures.ThreadPoolExecutor`. The maximum number of
    threads available is controlled by the ``max_concurrent`` argument.

    .. warning::
        Normally the processing of each Request occurs in a distinct thread context. If
        you need to access shared state please be careful to use appropriate
        concurrency mechanisms.

    .. warning::
        The default value for ``max_concurrent`` is 5, but setting it to 1 is allowed.
        This means that a Plugin will essentially be single-threaded, but realize this
        means that if the Plugin invokes a Command on itself in the course of processing
        a Request then the Plugin **will** deadlock!

    Args:
        client: Instance of a class annotated with ``@system``.

        bg_host (str): Beer-garden hostname
        bg_port (int): Beer-garden port
        bg_url_prefix (str): URL path that will be used as a prefix when communicating
            with Beer-garden. Useful if Beer-garden is running on a URL other than '/'.
        ssl_enabled (bool): Whether to use SSL for Beer-garden communication
        ca_cert (str): Path to certificate file containing the certificate of the
            authority that issued the Beer-garden server certificate
        ca_verify (bool): Whether to verify Beer-garden server certificate
        client_cert (str): Path to client certificate to use when communicating with
            Beer-garden
        api_version (int): Beer-garden API version to use
        client_timeout (int): Max time to wait for Beer-garden server response
        username (str): Username for Beer-garden authentication
        password (str): Password for Beer-garden authentication
        access_token (str): Access token for Beer-garden authentication
        refresh_token (str): Refresh token for Beer-garden authentication

        system (:py:class:`brewtils.models.System`): A Beer-garden System definition.
            Incompatible with name, version, description, display_name, icon_name,
            max_instances, and metadata parameters.
        name (str): System name
        version (str): System version
        description (str): System description
        display_name (str): System display name
        icon_name (str): System icon name
        max_instances (int): System maximum instances
        metadata (dict): System metadata
        instance_name (str): Instance name
        namespace (str): Namespace name

        logger (:py:class:`logging.Logger`): Logger that will be used by the Plugin.
            Passing a logger will prevent the Plugin from preforming any additional
            logging configuration.

        worker_shutdown_timeout (int): Time to wait during shutdown to finish processing
        max_concurrent (int): Maximum number of requests to process concurrently
        max_attempts (int): Number of times to attempt updating of a Request
            before giving up. Negative numbers are interpreted as no maximum.
        max_timeout (int): Maximum amount of time to wait between Request update
            attempts. Negative numbers are interpreted as no maximum.
        starting_timeout (int): Initial time to wait between Request update attempts.
            Will double on subsequent attempts until reaching max_timeout.

        mq_max_attempts (int): Number of times to attempt reconnection to message queue
            before giving up. Negative numbers are interpreted as no maximum.
        mq_max_timeout (int): Maximum amount of time to wait between message queue
            reconnect attempts. Negative numbers are interpreted as no maximum.
        mq_starting_timeout (int): Initial time to wait between message queue reconnect
            attempts. Will double on subsequent attempts until reaching mq_max_timeout.
        working_directory (str): Path to a preferred working directory. Only used
            when working with bytes parameters.
    """

    def __init__(self, client=None, system=None, logger=None, **kwargs):
        self._client = None
        self._instance = None
        self._admin_processor = None
        self._request_processor = None
        self._shutdown_event = threading.Event()

        # Need to set up logging before loading config
        self._custom_logger = False
        self._logger = self._setup_logging(logger=logger, **kwargs)

        # Now that logging is configured we can load the real config
        self._config = load_config(**kwargs)

        # If global config has already been set that's a warning
        global CONFIG
        if len(CONFIG):
            self._logger.warning(
                "Global CONFIG object is not empty! If multiple plugins are running in "
                "this process please ensure any [System|Easy|Rest]Clients are passed "
                "connection information as kwargs as auto-discovery may be incorrect."
            )
        CONFIG = Box(self._config.to_dict(), default_box=True)

        # Now set up the system
        self._system = self._setup_system(system, kwargs)

        # Make sure this is set after self._system
        if client:
            self.client = client

        # Now that the config is loaded we can create the EasyClient
        self._ez_client = EasyClient(logger=self._logger, **self._config)

        # With the EasyClient we can determine if this is an old garden
        self._legacy = self._legacy_garden()

        if not self._legacy:
            # Namespace setup depends on self._system and self._ez_client
            self._setup_namespace()

            # And with _system and _ez_client we can ask for the real logging config
            self._initialize_logging()

    def run(self):
        if not self._client:
            raise AttributeError(
                "Unable to start a Plugin without a Client. Please set the 'client' "
                "attribute to an instance of a class decorated with @brewtils.system"
            )

        self._startup()
        self._logger.info("Plugin %s has started", self.unique_name)

        try:
            # Need the timeout param so this works correctly in Python 2
            while not self._shutdown_event.wait(timeout=0.1):
                pass
        except KeyboardInterrupt:
            self._logger.debug("Received KeyboardInterrupt - shutting down")
        except Exception as ex:
            self._logger.exception("Exception during wait, shutting down: %s", ex)

        self._shutdown()
        self._logger.info("Plugin %s has terminated", self.unique_name)

    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, new_client):
        if self._client:
            raise AttributeError("Sorry, you can't change a plugin's client once set")

        if new_client is None:
            return

        # Several _system properties can come from the client, so update if needed
        if not self._system.name:
            self._system.name = getattr(new_client, "_bg_name")
        if not self._system.version:
            self._system.version = getattr(new_client, "_bg_version")
        if not self._system.description and new_client.__doc__:
            self._system.description = new_client.__doc__.split("\n")[0]

        # Now roll up / interpret all metadata to get the Commands
        self._system.commands = _parse_client(new_client)

        try:
            # Put some attributes on the Client class
            client_clazz = type(new_client)
            client_clazz.current_request = property(
                lambda _: request_context.current_request
            )

            # Add for back-compatibility
            client_clazz._bg_name = self._system.name
            client_clazz._bg_version = self._system.version
            client_clazz._bg_commands = self._system.commands
            client_clazz._current_request = client_clazz.current_request
        except TypeError:
            if sys.version_info.major != 2:
                raise

            self._logger.warning(
                "Unable to assign attributes to Client class - current_request will "
                "not be available. If you're using an old-style class declaration "
                "it's recommended to switch to new-style if possible."
            )

        self._client = new_client

    @property
    def system(self):
        return self._system

    @property
    def instance(self):
        return self._instance

    @property
    def unique_name(self):
        return "%s:%s[%s]-%s" % (
            self._system.namespace,
            self._system.name,
            self._config.instance_name,
            self._system.version,
        )

    @staticmethod
    def _set_signal_handlers():
        """Ensure that SIGINT and SIGTERM will gracefully stop the Plugin"""

        def _handler(_signal, _frame):
            raise KeyboardInterrupt

        signal.signal(signal.SIGINT, _handler)
        signal.signal(signal.SIGTERM, _handler)

    @staticmethod
    def _set_exception_hook(logger):
        """Ensure uncaught exceptions are logged instead of being written to stderr"""

        def _hook(exc_type, exc_value, traceback):
            logger.error(
                "An uncaught exception was raised in the plugin process:",
                exc_info=(exc_type, exc_value, traceback),
            )

        sys.excepthook = _hook

    def _startup(self):
        """Plugin startup procedure

        This method actually starts the plugin. When it completes the plugin should be
        considered in a "running" state - listening to the appropriate message queues,
        connected to the Beer-garden server, and ready to process Requests.

        This method should be the first time that a connection to the Beer-garden
        server is *required*.
        """
        self._logger.debug("About to start up plugin %s", self.unique_name)

        if not self._ez_client.can_connect():
            raise RestConnectionError("Cannot connect to the Beer-garden server")

        # If namespace couldn't be determined at init try one more time
        if not self._legacy and not self._config.namespace:
            self._setup_namespace()

        self._system = self._initialize_system()
        self._instance = self._initialize_instance()

        if self._config.working_directory is None:
            app_parts = [self._system.name, self._instance.name]
            if self._system.namespace:
                app_parts.insert(0, self._system.namespace)

            self._config.working_directory = appdirs.user_data_dir(
                appname=os.path.join(*app_parts), version=self._system.version
            )

            workdir = Path(self._config.working_directory)
            if not workdir.exists():
                workdir.mkdir(parents=True)

        self._logger.debug("Initializing and starting processors")
        self._admin_processor, self._request_processor = self._initialize_processors()
        self._admin_processor.startup()
        self._request_processor.startup()

        self._logger.debug("Setting signal handlers")
        self._set_signal_handlers()

    def _shutdown(self):
        """Plugin shutdown procedure

        This method gracefully stops the plugin. When it completes the plugin should be
        considered in a "stopped" state - the message processors shut down and all
        connections closed.
        """
        self._logger.debug("About to shut down plugin %s", self.unique_name)
        self._shutdown_event.set()

        self._logger.debug("Shutting down processors")
        self._request_processor.shutdown()
        self._admin_processor.shutdown()

        try:
            self._ez_client.update_instance(self._instance.id, new_status="STOPPED")
        except Exception:
            self._logger.warning(
                "Unable to notify Beer-garden that this plugin is STOPPED, so this "
                "plugin's status may be incorrect in Beer-garden"
            )

        self._logger.debug("Successfully shutdown plugin {0}".format(self.unique_name))

    def _initialize_logging(self):
        """Configure logging with Beer-garden's configuration for this plugin.

        This method will ask Beer-garden for a logging configuration specific to this
        plugin and will apply that configuration to the logging module.

        Note that this method will do nothing if the logging module's configuration was
        already set or a logger kwarg was given during Plugin construction.

        Returns:
            None
        """
        if self._custom_logger:
            self._logger.debug("Skipping logging init: custom logger detected")
            return

        try:
            log_config = self._ez_client.get_logging_config(
                local=bool(self._config.runner_id)
            )
        except Exception as ex:
            self._logger.warning(
                "Unable to retrieve logging configuration from Beergarden, the default "
                "configuration will be used instead. Caused by: {0}".format(ex)
            )
            return

        try:
            configure_logging(
                log_config,
                namespace=self._system.namespace,
                system_name=self._system.name,
                system_version=self._system.version,
                instance_name=self._config.instance_name,
            )
        except Exception as ex:
            # Reset to default config as logging can be seriously wrong now
            logging.config.dictConfig(default_config(level=self._config.log_level))

            self._logger.exception(
                "Error encountered during logging configuration. This most likely "
                "indicates an issue with the Beergarden server plugin logging "
                "configuration. The default configuration will be used instead. Caused "
                "by: {0}".format(ex)
            )
            return

        # Finally, log uncaught exceptions using the configuration instead of stderr
        self._set_exception_hook(self._logger)

    def _initialize_system(self):
        """Let Beergarden know about System-level info

        This will attempt to find a system with a name and version matching this plugin.
        If one is found this will attempt to update it (with commands, metadata, etc.
        from this plugin).

        If a System is not found this will attempt to create one.

        Returns:
            Definition of a Beergarden System this plugin belongs to.

        Raises:
            PluginValidationError: Unable to find or create a System for this Plugin
        """
        # Make sure that the system is actually valid before trying anything
        self._validate_system()

        # Do any necessary template resolution
        self._system.template = resolve_template(self._system.template)

        existing_system = self._ez_client.find_unique_system(
            name=self._system.name,
            version=self._system.version,
            namespace=self._system.namespace,
        )

        if not existing_system:
            try:
                # If this succeeds can just finish here
                return self._ez_client.create_system(self._system)
            except ConflictError:
                # If multiple instances are starting up at once and this is a new system
                # the create can return a conflict. In that case just try the get again
                existing_system = self._ez_client.find_unique_system(
                    name=self._system.name,
                    version=self._system.version,
                    namespace=self._system.namespace,
                )

        # If we STILL can't find a system something is really wrong
        if not existing_system:
            raise PluginValidationError(
                "Unable to find or create system {0}".format(self._system)
            )

        # We always update with these fields
        update_kwargs = {
            "new_commands": self._system.commands,
            "metadata": self._system.metadata,
            "description": self._system.description,
            "display_name": self._system.display_name,
            "icon_name": self._system.icon_name,
            "template": self._system.template,
        }

        # And if this particular instance doesn't exist we want to add it
        if not existing_system.has_instance(self._config.instance_name):
            update_kwargs["add_instance"] = Instance(name=self._config.instance_name)

        return self._ez_client.update_system(existing_system.id, **update_kwargs)

    def _initialize_instance(self):
        """Let Beer-garden know this instance is ready to process Requests"""
        # Sanity check to make sure an instance with this name was registered
        if not self._system.has_instance(self._config.instance_name):
            raise PluginValidationError(
                "Unable to find registered instance with name '%s'"
                % self._config.instance_name
            )

        return self._ez_client.initialize_instance(
            self._system.get_instance_by_name(self._config.instance_name).id,
            runner_id=self._config.runner_id,
        )

    def _initialize_processors(self):
        """Create RequestProcessors for the admin and request queues"""
        # If the queue connection is TLS we need to update connection params with
        # values specified at plugin creation
        connection_info = self._instance.queue_info["connection"]
        if "ssl" in connection_info:
            connection_info["ssl"].update(
                {
                    "ca_cert": self._config.ca_cert,
                    "ca_verify": self._config.ca_verify,
                    "client_cert": self._config.client_cert,
                }
            )

        # Each RequestProcessor needs a RequestConsumer, so start with those
        common_args = {
            "connection_type": self._instance.queue_type,
            "connection_info": connection_info,
            "panic_event": self._shutdown_event,
            "max_reconnect_attempts": self._config.mq.max_attempts,
            "max_reconnect_timeout": self._config.mq.max_timeout,
            "starting_reconnect_timeout": self._config.mq.starting_timeout,
        }
        admin_consumer = RequestConsumer.create(
            thread_name="Admin Consumer",
            queue_name=self._instance.queue_info["admin"]["name"],
            max_concurrent=1,
            **common_args
        )
        request_consumer = RequestConsumer.create(
            thread_name="Request Consumer",
            queue_name=self._instance.queue_info["request"]["name"],
            max_concurrent=self._config.max_concurrent,
            **common_args
        )

        # Both RequestProcessors need an updater
        updater = HTTPRequestUpdater(
            self._ez_client,
            self._shutdown_event,
            max_attempts=self._config.max_attempts,
            max_timeout=self._config.max_timeout,
            starting_timeout=self._config.starting_timeout,
        )

        # Finally, create the actual RequestProcessors
        admin_processor = AdminProcessor(
            target=self,
            updater=updater,
            consumer=admin_consumer,
            plugin_name=self.unique_name,
            max_workers=1,
        )
        request_processor = RequestProcessor(
            target=self._client,
            updater=updater,
            consumer=request_consumer,
            validation_funcs=[self._correct_system, self._is_running],
            plugin_name=self.unique_name,
            max_workers=self._config.max_concurrent,
            resolver=ResolutionManager(easy_client=self._ez_client),
            system=self._system,
        )

        return admin_processor, request_processor

    def _start(self):
        """Handle start Request"""
        self._instance = self._ez_client.update_instance(
            self._instance.id, new_status="RUNNING"
        )

    def _stop(self):
        """Handle stop Request"""
        # Because the run() method is on a 0.1s sleep there's a race regarding if the
        # admin consumer will start processing the next message on the queue before the
        # main thread can stop it. So stop it here to prevent that.
        self._request_processor.consumer.stop_consuming()
        self._admin_processor.consumer.stop_consuming()

        self._shutdown_event.set()

    def _status(self):
        """Handle status Request"""
        try:
            self._ez_client.instance_heartbeat(self._instance.id)
        except (RequestsConnectionError, RestConnectionError):
            pass

    def _read_log(self, **kwargs):
        """Handle read log Request"""

        log_file = find_log_file()

        if not log_file:
            raise RequestProcessingError(
                "Error attempting to retrieve logs - unable to determine log filename. "
                "Please verify that the plugin is writing to a log file."
            )

        try:
            return read_log_file(log_file=log_file, **kwargs)
        except IOError as e:
            raise RequestProcessingError(
                "Error attempting to retrieve logs - unable to read log file at {0}. "
                "Root cause I/O error {1}: {2}".format(log_file, e.errno, e.strerror)
            )

    def _correct_system(self, request):
        """Validate that a request is intended for this Plugin"""
        request_system = getattr(request, "system") or ""
        if request_system.upper() != self._system.name.upper():
            raise DiscardMessageException(
                "Received message for system {0}".format(request.system)
            )

    def _is_running(self, _):
        """Validate that this plugin is still running"""
        if self._shutdown_event.is_set():
            raise RequestProcessingError(
                "Unable to process message - currently shutting down"
            )

    def _legacy_garden(self):
        """Determine if this plugin is connected to a legacy garden"""
        legacy = False

        try:
            # Need to be careful since v2 doesn't have "beer_garden_version"
            raw_version = self._ez_client.get_version()
            if "beer_garden_version" in raw_version:
                bg_version = Version(raw_version["beer_garden_version"])
            else:
                bg_version = Version(raw_version["brew_view_version"])

            if bg_version < Version("3"):
                legacy = True

                _deprecate(
                    "Looks like your plugin is using version 3 brewtils but connecting "
                    "to a version 2 Beer Garden. Please be aware that this "
                    "functionality will stop being officially supported in the next "
                    "brewtils minor release."
                )

                self._logger.warning(
                    "This plugin is using brewtils version {0} but is connected to a "
                    "legacy Beer Garden (version {1}). Please be aware that certain "
                    "features such as namespaces and logging configuration will not "
                    "work correctly until the Beer Garden is upgraded.".format(
                        brewtils.__version__, bg_version
                    )
                )

        except Exception as ex:
            self._logger.warning(
                "An exception was raised while attempting to determine Beer Garden "
                "version, assuming non-legacy."
            )
            self._logger.debug("Underlying exception: %s" % ex, exc_info=True)

        return legacy

    def _setup_logging(self, logger=None, **kwargs):
        """Set up logging configuration and get a logger for the Plugin

        This method will configure Python-wide logging for the process if it has not
        already been configured. Whether or not logging has been configured is
        determined by the root handler count - if there aren't any then it's assumed
        logging has not already been configured.

        The configuration applied (again, if no configuration has already happened) is
        a stream handler with elevated log levels for libraries that are verbose. The
        overall level will be loaded as a configuration option, so it can be set as a
        keyword argument, command line option, or environment variable.

        A logger to be used by the Plugin will be returned. If the ``logger`` keyword
        parameter is given then that logger will be used, otherwise a logger will be
        generated from the standard ``logging`` module.

        Finally, if a the ``logger`` keyword parameter is supplied it's assumed that
        logging is already configured and no further configuration will be applied.

        Args:
            logger: A custom logger
            **kwargs: Will be used to load the bootstrap config

        Returns:
            A logger for the Plugin
        """
        if logger or len(logging.root.handlers) != 0:
            self._custom_logger = True
        else:
            # log_level is the only bootstrap config item
            boot_config = load_config(bootstrap=True, **kwargs)
            logging.config.dictConfig(default_config(level=boot_config.log_level))

            self._custom_logger = False

        return logger or logging.getLogger(__name__)

    def _setup_namespace(self):
        """Determine the namespace the Plugin is operating in

        This function attempts to determine the correct namespace and ensures that
        the value is set in the places it needs to be set.

        First, look in the resolved system (self._system) to see if that has a
        namespace. If it does, either:

        - A complete system definition with a namespace was provided
        - The namespace was resolved from the config

        In the latter case nothing further needs to be done. In the former case we
        need to set the global config namespace value to the system's namespace value
        so that any SystemClients created after the plugin will have the correct value.
        Because we have no way to know which case is correct we assume the former and
        always set the config value.

        If the system does not have a namespace then we attempt to use the EasyClient to
        determine the "default" namespace. If successful we set both the global config
        and the system namespaces to the default value.

        If the attempt to determine the default namespace is not successful we log a
        warning. We don't really want to *require* the connection to Beer-garden until
        Plugin.run() is called. Raising an exception here would do that, so instead we
        just log the warning. Another attempt will be made to determine the namespace
        in Plugin.run() which will raise on failure (but again, SystemClients created
        before the namespace is determined will have potentially incorrect namespaces).
        """
        try:
            ns = self._system.namespace or self._ez_client.get_config()["garden_name"]

            self._system.namespace = ns
            self._config.namespace = ns
            CONFIG.namespace = ns
        except Exception as ex:
            self._logger.warning(
                "Namespace value was not resolved from config sources and an exception "
                "was raised while attempting to determine default namespace value. "
                "Created SystemClients may have unexpected namespace values. "
                "Underlying exception was:\n%s" % ex
            )

    def _setup_system(self, system, plugin_kwargs):
        helper_keywords = {
            "name",
            "version",
            "description",
            "icon_name",
            "display_name",
            "max_instances",
            "metadata",
            "namespace",
            "template",
        }

        if system:
            if helper_keywords.intersection(plugin_kwargs.keys()):
                raise ValidationError(
                    "Sorry, you can't provide a complete system definition as well as "
                    "system creation helper kwargs %s" % helper_keywords
                )

            if not system.instances:
                raise ValidationError(
                    "Explicit system definition requires explicit instance "
                    "definition (use instances=[Instance(name='default')] for "
                    "default behavior)"
                )

            if not system.max_instances:
                system.max_instances = len(system.instances)

        else:
            # Commands are not defined here - they're set in the client property setter
            system = System(
                name=self._config.name,
                version=self._config.version,
                description=self._config.description,
                namespace=self._config.namespace,
                metadata=json.loads(self._config.metadata),
                instances=[Instance(name=self._config.instance_name)],
                max_instances=self._config.max_instances,
                icon_name=self._config.icon_name,
                display_name=self._config.display_name,
                template=self._config.template,
            )

        return system

    def _validate_system(self):
        """Make sure the System definition makes sense"""
        if not self._system.name:
            raise ValidationError("Plugin system must have a name")

        if not self._system.version:
            raise ValidationError("Plugin system must have a version")

        client_name = getattr(self._client, "_bg_name", None)
        if client_name and client_name != self._system.name:
            raise ValidationError(
                "System name '%s' doesn't match name from client decorator: "
                "@system(bg_name=%s)" % (self._system.name, client_name)
            )

        client_version = getattr(self._client, "_bg_version", None)
        if client_version and client_version != self._system.version:
            raise ValidationError(
                "System version '%s' doesn't match version from client decorator: "
                "@system(bg_version=%s)" % (self._system.version, client_version)
            )

    # These are provided for backward-compatibility
    @property
    def bg_host(self):
        """
        .. deprecated:: 3.0
           bg_host is now in ``_config`` (``plugin._config.bg_host``)

        Provided for backward-comptibility
        """
        _deprecate("bg_host is now in _config (plugin._config.bg_host)")
        return self._config.bg_host

    @property
    def bg_port(self):
        """
        .. deprecated:: 3.0
           bg_port is now in _config (``plugin._config.bg_port``)

        Provided for backward-comptibility
        """
        _deprecate("bg_port is now in _config (plugin._config.bg_port)")
        return self._config.bg_port

    @property
    def ssl_enabled(self):
        """
        .. deprecated:: 3.0
           ssl_enabled is now in ``_config`` (``plugin._config.ssl_enabled``)

        Provided for backward-comptibility
        """
        _deprecate("ssl_enabled is now in _config (plugin._config.ssl_enabled)")
        return self._config.ssl_enabled

    @property
    def ca_cert(self):
        """
        .. deprecated:: 3.0
           ca_cert is now in ``_config`` (``plugin._config.ca_cert``)

        Provided for backward-comptibility
        """
        _deprecate("ca_cert is now in _config (plugin._config.ca_cert)")
        return self._config.ca_cert

    @property
    def client_cert(self):
        """
        .. deprecated:: 3.0
           client_cert is now in ``_config`` (``plugin._config.client_cert``)

        Provided for backward-comptibility
        """
        _deprecate("client_cert is now in _config (plugin._config.client_cert)")
        return self._config.client_cert

    @property
    def bg_url_prefix(self):
        """
        .. deprecated:: 3.0
           bg_url_prefix is now in ``_config`` (``plugin._config.bg_url_prefix``)

        Provided for backward-comptibility
        """
        _deprecate("bg_url_prefix is now in _config (plugin._config.bg_url_prefix)")
        return self._config.bg_url_prefix

    @property
    def ca_verify(self):
        """
        .. deprecated:: 3.0
           ca_verify is now in ``_config`` (``plugin._config.ca_verify``)

        Provided for backward-comptibility
        """
        _deprecate("ca_verify is now in _config (plugin._config.ca_verify)")
        return self._config.ca_verify

    @property
    def max_attempts(self):
        """
        .. deprecated:: 3.0
           max_attempts is now in ``_config`` (``plugin._config.max_attempts``)

        Provided for backward-comptibility
        """
        _deprecate("max_attempts is now in _config (plugin._config.max_attempts)")
        return self._config.max_attempts

    @property
    def max_timeout(self):
        """
        .. deprecated:: 3.0
           max_timeout is now in ``_config`` (``plugin._config.max_timeout``)

        Provided for backward-comptibility
        """
        _deprecate("max_timeout has moved into _config (plugin._config.max_timeout)")
        return self._config.max_timeout

    @property
    def starting_timeout(self):
        """
        .. deprecated:: 3.0
           starting_timeout is now in ``_config`` (``plugin._config.starting_timeout``)

        Provided for backward-comptibility
        """
        _deprecate(
            "starting_timeout is now in _config (plugin._config.starting_timeout)"
        )
        return self._config.starting_timeout

    @property
    def max_concurrent(self):
        """
        .. deprecated:: 3.0
           max_concurrent is now in ``_config`` (``plugin._config.max_concurrent``)

        Provided for backward-comptibility
        """
        _deprecate("max_concurrent is now in _config (plugin._config.max_concurrent)")
        return self._config.max_concurrent

    @property
    def instance_name(self):
        """
        .. deprecated:: 3.0
           instance_name is now in ``_config`` (``plugin._config.instance_name``)

        Provided for backward-comptibility
        """
        _deprecate("instance_name is now in _config (plugin._config.instance_name)")
        return self._config.instance_name

    @property
    def connection_parameters(self):
        """
        .. deprecated:: 3.0
           connection_parameters has been removed. Please use ``_config``

        Provided for backward-comptibility
        """
        _deprecate("connection_parameters attribute was removed, please use '_config'")
        return {key: self._config[key] for key in _CONNECTION_SPEC}

    @property
    def metadata(self):
        """
        .. deprecated:: 3.0
           metadata is now part of the ``system`` attribute (``plugin.system.metadata``)

        Provided for backward-comptibility
        """
        _deprecate("metadata is a part of the system attribute (plugin.system.metadata")
        return self._system.metadata

    @property
    def bm_client(self):
        """
        .. deprecated:: 3.0
            bm_client attribute has been renamed to ``_ez_client``.

        Provided for backward-comptibility
        """
        _deprecate("bm_client attribute has been renamed to _ez_client")
        return self._ez_client

    @property
    def shutdown_event(self):
        """
        .. deprecated:: 3.0
            shutdown_event attribute has been renamed to ``_shutdown_event``.

        Provided for backward-comptibility
        """
        _deprecate("shutdown_event attribute has been renamed to _shutdown_event")
        return self._shutdown_event

    @property
    def logger(self):
        """
        .. deprecated:: 3.0
            logger attribute has been renamed to ``_logger``.

        Provided for backward-comptibility
        """
        _deprecate("logger attribute has been renamed to _logger")
        return self._logger


# Alias old names
class PluginBase(Plugin):
    """
    .. deprecated:: 3.0
        Will be removed in version 4.0. Please use ``Plugin`` instead.

    ``Plugin`` alias Provided for backward-comptibility
    """

    def __init__(self, *args, **kwargs):
        _deprecate(
            "Looks like you're creating a 'PluginBase'. Heads up - this name will be "
            "removed in version 4.0, please use 'Plugin' instead. Thanks!"
        )
        super(PluginBase, self).__init__(*args, **kwargs)


class RemotePlugin(Plugin):
    """
    .. deprecated:: 3.0
        Will be removed in version 4.0. Please use ``Plugin`` instead.

    ``Plugin`` alias Provided for backward-comptibility
    """

    def __init__(self, *args, **kwargs):
        _deprecate(
            "Looks like you're creating a 'RemotePlugin'. Heads up - this name will be "
            "removed in version 4.0, please use 'Plugin' instead. Thanks!"
        )
        super(RemotePlugin, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-

import functools
import inspect
import os
import sys
from types import MethodType
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, Union

import six

from brewtils.choices import process_choices
from brewtils.display import resolve_form, resolve_schema, resolve_template
from brewtils.errors import PluginParamError, _deprecate
from brewtils.models import Command, Parameter, Resolvable

if sys.version_info.major == 2:
    from funcsigs import signature, Parameter as InspectParameter  # noqa
else:
    from inspect import signature, Parameter as InspectParameter  # noqa

__all__ = [
    "client",
    "command",
    "parameter",
    "parameters",
    "system",
]


def client(
    _wrapped=None,  # type: Type
    bg_name=None,  # type: Optional[str]
    bg_version=None,  # type: Optional[str]
):
    # type: (...) -> Type
    """Class decorator that marks a class as a beer-garden Client

    Using this decorator is no longer strictly necessary. It was previously required in
    order to mark a class as being a Beer-garden Client, and contained most of the logic
    that currently resides in the ``parse_client`` function. However, that's no longer
    the case and this currently exists mainly for back-compatibility reasons.

    Applying this decorator to a client class does have the nice effect of preventing
    linters from complaining if any special attributes are used. So that's something.

    Those special attributes are below. Note that these are just placeholders until the
    actual values are populated when the client instance is assigned to a Plugin:

      * ``_bg_name``: an optional system name
      * ``_bg_version``: an optional system version
      * ``_bg_commands``: holds all registered commands
      * ``_current_request``: Reference to the currently executing request

    Args:
        _wrapped: The class to decorate. This is handled as a positional argument and
            shouldn't be explicitly set.
        bg_name: Optional plugin name
        bg_version: Optional plugin version

    Returns:
        The decorated class

    """
    if _wrapped is None:
        return functools.partial(client, bg_name=bg_name, bg_version=bg_version)  # noqa

    # Assign these here so linters don't complain
    _wrapped._bg_name = bg_name
    _wrapped._bg_version = bg_version
    _wrapped._bg_commands = []
    _wrapped._current_request = None

    return _wrapped


def command(
    _wrapped=None,  # type: Union[Callable, MethodType]
    description=None,  # type: Optional[str]
    parameters=None,  # type: Optional[List[Parameter]]
    command_type="ACTION",  # type: str
    output_type="STRING",  # type: str
    schema=None,  # type: Optional[Union[dict, str]]
    form=None,  # type: Optional[Union[dict, list, str]]
    template=None,  # type: Optional[str]
    icon_name=None,  # type: Optional[str]
    hidden=False,  # type: Optional[bool]
    metadata=None,  # type: Optional[Dict]
):
    """Decorator for specifying Command details

    For example:

    .. code-block:: python

        @command(output_type='JSON')
        def echo_json(self, message):
            return message

    Args:
        _wrapped: The function to decorate. This is handled as a positional argument and
            shouldn't be explicitly set.
        description: The command description. If not given the first line of the method
            docstring will be used.
        parameters: A list of Command parameters. It's recommended to use @parameter
            decorators to declare Parameters instead of declaring them here, but it is
            allowed. Any Parameters given here will be merged with Parameters sourced
            from decorators and inferred from the method signature.
        command_type: The command type. Valid options are Command.COMMAND_TYPES.
        output_type: The output type. Valid options are Command.OUTPUT_TYPES.
        schema: A custom schema definition.
        form: A custom form definition.
        template: A custom template definition.
        icon_name: The icon name. Should be either a FontAwesome or a Glyphicon name.
        hidden: Flag controlling whether the command is visible on the user interface.
        metadata: Free-form dictionary

    Returns:
        The decorated function
    """
    if _wrapped is None:
        return functools.partial(
            command,
            description=description,
            parameters=parameters,
            command_type=command_type,
            output_type=output_type,
            schema=schema,
            form=form,
            template=template,
            icon_name=icon_name,
            hidden=hidden,
            metadata=metadata,
        )

    new_command = Command(
        description=description,
        parameters=parameters,
        command_type=command_type,
        output_type=output_type,
        schema=schema,
        form=form,
        template=template,
        icon_name=icon_name,
        hidden=hidden,
        metadata=metadata,
    )

    # Python 2 compatibility
    if hasattr(_wrapped, "__func__"):
        _wrapped.__func__._command = new_command
    else:
        _wrapped._command = new_command

    return _wrapped


def parameter(
    _wrapped=None,  # type: Union[Callable, MethodType, Type]
    key=None,  # type: str
    type=None,  # type: Optional[Union[str, Type]]
    multi=None,  # type: Optional[bool]
    display_name=None,  # type: Optional[str]
    optional=None,  # type: Optional[bool]
    default=None,  # type: Optional[Any]
    description=None,  # type: Optional[str]
    choices=None,  # type: Optional[Union[Callable, Dict, Iterable, str]]
    parameters=None,  # type: Optional[List[Parameter]]
    nullable=None,  # type: Optional[bool]
    maximum=None,  # type: Optional[int]
    minimum=None,  # type: Optional[int]
    regex=None,  # type: Optional[str]
    form_input_type=None,  # type: Optional[str]
    type_info=None,  # type: Optional[dict]
    is_kwarg=None,  # type: Optional[bool]
    model=None,  # type: Optional[Type]
):
    """Decorator for specifying Parameter details

    For example::

        @parameter(
            key="message",
            description="Message to echo",
            optional=True,
            type="String",
            default="Hello, World!",
        )
        def echo(self, message):
            return message

    Args:
        _wrapped: The function to decorate. This is handled as a positional argument and
            shouldn't be explicitly set.
        key: String specifying the parameter identifier. If the decorated object is a
            method the key must match an argument name.
        type: String indicating the type to use for this parameter.
        multi: Boolean indicating if this parameter is a multi. See documentation for
            discussion of what this means.
        display_name: String that will be displayed as a label in the user interface.
        optional: Boolean indicating if this parameter must be specified.
        default: The value this parameter will be assigned if not overridden when
            creating a request.
        description: An additional string that will be displayed in the user interface.
        choices: List or dictionary specifying allowed values. See documentation for
            more information.
        parameters: Any nested parameters. See also: the 'model' argument.
        nullable: Boolean indicating if this parameter is allowed to be null.
        maximum: Integer indicating the maximum value of the parameter.
        minimum: Integer indicating the minimum value of the parameter.
        regex: String describing a regular expression constraint on the parameter.
        form_input_type: Specify the form input field type (e.g. textarea). Only used
            for string fields.
        type_info: Type-specific information. Mostly reserved for future use.
        is_kwarg: Boolean indicating if this parameter is meant to be part of the
            decorated function's kwargs. Only applies when the decorated object is a
            method.
        model: Class to be used as a model for this parameter. Must be a Python type
            object, not an instance.

    Returns:
        The decorated function

    """
    if _wrapped is None:
        return functools.partial(
            parameter,
            key=key,
            type=type,
            multi=multi,
            display_name=display_name,
            optional=optional,
            default=default,
            description=description,
            choices=choices,
            parameters=parameters,
            nullable=nullable,
            maximum=maximum,
            minimum=minimum,
            regex=regex,
            form_input_type=form_input_type,
            type_info=type_info,
            is_kwarg=is_kwarg,
            model=model,
        )

    new_parameter = Parameter(
        key=key,
        type=type,
        multi=multi,
        display_name=display_name,
        optional=optional,
        default=default,
        description=description,
        choices=choices,
        parameters=parameters,
        nullable=nullable,
        maximum=maximum,
        minimum=minimum,
        regex=regex,
        form_input_type=form_input_type,
        type_info=type_info,
        is_kwarg=is_kwarg,
        model=model,
    )

    # Python 2 compatibility
    if hasattr(_wrapped, "__func__"):
        _wrapped.__func__.parameters = getattr(_wrapped, "parameters", [])
        _wrapped.__func__.parameters.insert(0, new_parameter)
    else:
        _wrapped.parameters = getattr(_wrapped, "parameters", [])
        _wrapped.parameters.insert(0, new_parameter)

    return _wrapped


def parameters(*args, **kwargs):
    """
    .. deprecated:: 3.0
        Will be removed in version 4.0. Please use ``@command`` instead.

    Decorator for specifying multiple Parameter definitions at once

    This can be useful for commands which have a large number of complicated
    parameters but aren't good candidates for a Model.

    .. code-block:: python

        @parameter(**params[cmd1][param1])
        @parameter(**params[cmd1][param2])
        @parameter(**params[cmd1][param3])
        def cmd1(self, **kwargs):
            pass

    Can become:

    .. code-block:: python

        @parameters(params[cmd1])
        def cmd1(self, **kwargs):
            pass

    Args:
        *args (iterable): Positional arguments
            The first (and only) positional argument must be a list containing
            dictionaries that describe parameters.
        **kwargs: Used for bookkeeping. Don't set any of these yourself!

    Returns:
        func: The decorated function
    """
    # This is the first invocation
    if not kwargs.get("_partial"):
        # Need the callable check to prevent applying the decorator with no parenthesis
        if len(args) == 1 and not callable(args[0]):
            return functools.partial(parameters, args[0], _partial=True)

        raise PluginParamError("@parameters takes a single argument")

    # This is the second invocation
    else:
        if len(args) != 2:
            raise PluginParamError(
                "Incorrect number of arguments for parameters partial call. Did you "
                "set _partial=True? If so, please don't do that. If not, please let "
                "the Beergarden team know how you got here!"
            )

    _deprecate(
        "Looks like you're using the '@parameters' decorator. This is now deprecated - "
        "for passing bulk parameter definitions it's recommended to use the @command "
        "decorator parameters kwarg, like this: @command(parameters=[...])"
    )

    params = args[0]
    _wrapped = args[1]

    if not callable(_wrapped):
        raise PluginParamError("@parameters must be applied to a callable")

    try:
        for param in params:
            parameter(_wrapped, **param)
    except TypeError:
        raise PluginParamError("@parameters arg must be an iterable of dictionaries")

    return _wrapped


def _parse_client(client):
    # type: (object) -> List[Command]
    """Get a list of Beergarden Commands from a client object

    This will iterate over everything returned from dir, looking for metadata added
    by the decorators.

    """
    bg_commands = []

    for attr in dir(client):
        method = getattr(client, attr)

        method_command = _parse_method(method)

        if method_command:
            bg_commands.append(method_command)

    return bg_commands


def _parse_method(method):
    # type: (MethodType) -> Optional[Command]
    """Parse a method object as a Beer-garden command target

    If the method looks like a valid command target (based on the presence of certain
    attributes) then this method will initialize things:

    - The command will be initialized.
    - Every parameter will be initialized. Initializing a parameter is recursive - each
      nested parameter will also be initialized.
    - Top-level parameters are validated to ensure they match the method signature.

    Args:
        method: Method to parse

    Returns:
        Beergarden Command targeting the given method
    """
    if (inspect.ismethod(method) or inspect.isfunction(method)) and (
        hasattr(method, "_command") or hasattr(method, "parameters")
    ):
        # Create a command object if there isn't one already
        method_command = _initialize_command(method)

        try:
            # Need to initialize existing parameters before attempting to add parameters
            # pulled from the method signature.
            method_command.parameters = _initialize_parameters(
                method_command.parameters + getattr(method, "parameters", [])
            )

            # Add and update parameters based on the method signature
            _signature_parameters(method_command, method)

            # Verify that all parameters conform to the method signature
            _signature_validate(method_command, method)

        except PluginParamError as ex:
            six.raise_from(
                PluginParamError(
                    "Error initializing parameters for command '%s': %s"
                    % (method_command.name, ex)
                ),
                ex,
            )

        return method_command


def _initialize_command(method):
    # type: (MethodType) -> Command
    """Initialize a Command

    This takes care of ensuring a Command object is in the correct form. Things like:

    - Assigning the name from the method name
    - Pulling the description from the method docstring, if necessary
    - Resolving display modifiers (schema, form, template)

    Args:
        method: The method with the Command to initialize

    Returns:
        The initialized Command

    """
    cmd = getattr(method, "_command", Command())

    cmd.name = _method_name(method)
    cmd.description = cmd.description or _method_docstring(method)

    try:
        base_dir = os.path.dirname(inspect.getfile(method))

        cmd.schema = resolve_schema(cmd.schema, base_dir=base_dir)
        cmd.form = resolve_form(cmd.form, base_dir=base_dir)
        cmd.template = resolve_template(cmd.template, base_dir=base_dir)
    except PluginParamError as ex:
        six.raise_from(
            PluginParamError("Error initializing command '%s': %s" % (cmd.name, ex)),
            ex,
        )

    return cmd


def _method_name(method):
    # type: (MethodType) -> str
    """Get the name of a method

    This is needed for Python 2 / 3 compatibility

    Args:
        method: Method to inspect

    Returns:
        Method name

    """
    if hasattr(method, "func_name"):
        command_name = method.func_name
    else:
        command_name = method.__name__

    return command_name


def _method_docstring(method):
    # type: (MethodType) -> str
    """Parse out the first line of the docstring from a method

    This is needed for Python 2 / 3 compatibility

    Args:
        method: Method to inspect

    Returns:
        First line of docstring

    """
    if hasattr(method, "func_doc"):
        docstring = method.func_doc
    else:
        docstring = method.__doc__

    return docstring.split("\n")[0] if docstring else None


def _sig_info(arg):
    # type: (InspectParameter) -> Tuple[Any, bool]
    """Get the default and optionality of a method argument

    This will return the "default" according to the method signature. For example, the
    following would return "foo" as the default for Parameter param:

    .. code-block:: python

        def my_command(self, param="foo"):
            ...

    The "optional" value returned will be a boolean indicating the presence of a default
    argument. In the example above the "optional" value will be True. However, in the
    following example the value would be False (and the "default" value will be None):

    .. code-block:: python

        def my_command(self, param):
            ...

    A separate optional return is needed to indicate when a default is provided in the
    signature, but the default is None. In the following, the default will still be
    None, but the optional value will be True:

    .. code-block:: python

        def my_command(self, param=None):
            ...

    Args:
        arg: The method argument

    Returns:
        Tuple of (signature default, optionality)
    """
    if arg.default != InspectParameter.empty:
        return arg.default, True
    return None, False


def _initialize_parameter(
    param=None,
    key=None,
    type=None,
    multi=None,
    display_name=None,
    optional=None,
    default=None,
    description=None,
    choices=None,
    parameters=None,
    nullable=None,
    maximum=None,
    minimum=None,
    regex=None,
    form_input_type=None,
    type_info=None,
    is_kwarg=None,
    model=None,
):
    # type: (...) -> Parameter
    """Initialize a Parameter

    This exists to move logic out of the @parameter decorator. Previously there was a
    fair amount of logic in the decorator, which meant that it wasn't feasible to create
    a Parameter without using it. This made things like nested models difficult to do
    correctly.

    There are also some checks and translation that need to happen for every Parameter,
    most notably the "choices" attribute.

    This method also ensures that these checks and translations occur for child
    Parameters.

    Args:
        param: An already-created Parameter. If this is given all the other
        Parameter-creation kwargs will be ignored

    Keyword Args:
        Will be used to construct a new Parameter
    """
    param = param or Parameter(
        key=key,
        type=type,
        multi=multi,
        display_name=display_name,
        optional=optional,
        default=default,
        description=description,
        choices=choices,
        parameters=parameters,
        nullable=nullable,
        maximum=maximum,
        minimum=minimum,
        regex=regex,
        form_input_type=form_input_type,
        type_info=type_info,
        is_kwarg=is_kwarg,
        model=model,
    )

    # Every parameter needs a key, so stop that right here
    if param.key is None:
        raise PluginParamError("Attempted to create a parameter without a key")

    # Type and type info
    # Type info is where type specific information goes. For now, this is specific
    # to file types. See #289 for more details.
    param.type = _format_type(param.type)
    param.type_info = param.type_info or {}
    if param.type in Resolvable.TYPES:
        param.type_info["storage"] = "gridfs"

        # Also nullify default parameters for safety
        param.default = None

    # Process the raw choices into a Choices object
    param.choices = process_choices(param.choices)

    # Now deal with nested parameters
    if param.parameters or param.model:
        if param.model:
            # Can't specify a model and parameters - which should win?
            if param.parameters:
                raise PluginParamError(
                    "Error initializing parameter '%s': A parameter with both a model "
                    "and nested parameters is not allowed" % param.key
                )

            param.parameters = param.model.parameters
            param.model = None

        param.type = "Dictionary"
        param.parameters = _initialize_parameters(param.parameters)

    return param


def _format_type(param_type):
    # type: (Any) -> str
    """Parse Parameter type

    Args:
        param_type: Raw Parameter type, usually from a decorator

    Returns:
        Properly formatted string describing the parameter type
    """
    if param_type == str:
        return "String"
    elif param_type == int:
        return "Integer"
    elif param_type == float:
        return "Float"
    elif param_type == bool:
        return "Boolean"
    elif param_type == dict:
        return "Dictionary"
    elif str(param_type).lower() == "datetime":
        return "DateTime"
    elif not param_type:
        return "Any"
    else:
        return str(param_type).title()


def _initialize_parameters(parameter_list):
    # type: (Iterable[Parameter, object, dict]) -> List[Parameter]
    """Initialize Parameters from a list of parameter definitions

    This exists for backwards compatibility with the old way of specifying Models.
    Previously, models were defined by creating a class with a ``parameters`` class
    attribute. This required constructing each parameter manually, without using the
    ``@parameter`` decorator.

    This function takes a list where members can be any of the following:
    - A Parameter object
    - A class object with a ``parameters`` attribute
    - A dict containing kwargs for constructing a Parameter

    The Parameters in the returned list will be initialized. See the function
    ``_initialize_parameter`` for information on what that entails.

    Args:
        parameter_list: List of parameter precursors

    Returns:
        List of initialized parameters
    """
    initialized_params = []

    for param in parameter_list:

        # This is already a Parameter. Only really need to interpret the choices
        # definition and recurse down into nested Parameters
        if isinstance(param, Parameter):
            initialized_params.append(_initialize_parameter(param=param))

        # This is a model class object. Needed for backwards compatibility
        # See https://github.com/beer-garden/beer-garden/issues/354
        elif hasattr(param, "parameters"):
            _deprecate(
                "Constructing a nested Parameters list using model class objects "
                "is deprecated. Please pass the model's parameter list directly."
            )
            initialized_params += _initialize_parameters(param.parameters)

        # This is a dict of Parameter kwargs
        elif isinstance(param, dict):
            initialized_params.append(_initialize_parameter(**param))

        # No clue!
        else:
            raise PluginParamError("Unable to generate parameter from '%s'" % param)

    return initialized_params


def _signature_parameters(cmd, method):
    # type: (Command, MethodType) -> Command
    """Add and/or modify a Command's parameters based on the method signature

    This will add / modify the Command's parameter list:

    - Any arguments in the method signature that were not already known Parameters will
      be added
    - Any arguments that WERE already known (most likely from a @parameter decorator)
      will potentially have their default and optional values updated:

      - If either attribute is already defined (specified in the decorator) then that
        value will be used. Explicit values will NOT be overridden.
      - If the default attribute is not already defined then it will be set to the value
        of the default parameter from the method signature, if any.
      - If the optional attribute is not already defined then it will be set to True if
        a default value exists in the method signature, otherwise it will be set to
        False.

    The parameter modification is confusing - see the _sig_info docstring for examples.

    A final note - I'm not super happy about this. It makes sense - positional arguments
    are "required", so mark them as non-optional. It's not *wrong*, but it's unexpected.
    A @parameter that doesn't specify "optional=" will have a different optionality
    based on the function signature.

    Regardless, we went with this originally. If we want to change it we need to go
    though a deprecation cycle and *loudly* publicize it since things wouldn't break
    loudly for plugin developers, their plugins would just be subtly (but importantly)
    different.

    Args:
        cmd: The Command to modify
        method: Method to parse

    Returns:
        Command with modified parameter list

    """
    # Now we need to reconcile the parameters with the method signature
    for index, arg in enumerate(signature(method).parameters.values()):

        # Don't want to include special parameters
        if (index == 0 and arg.name in ("self", "cls")) or arg.kind in (
            InspectParameter.VAR_KEYWORD,
            InspectParameter.VAR_POSITIONAL,
        ):
            continue

        # Grab default and optionality according to the signature. We'll need it later.
        sig_default, sig_optional = _sig_info(arg)

        # Here the parameter was not previously defined so just add it to the list
        if arg.name not in cmd.parameter_keys():
            cmd.parameters.append(
                _initialize_parameter(
                    key=arg.name, default=sig_default, optional=sig_optional
                )
            )

        # Here the parameter WAS previously defined. So we potentially need to update
        # the default and optional values (if they weren't explicitly set).
        else:
            param = cmd.get_parameter_by_key(arg.name)

            if param.default is None:
                param.default = sig_default

            if param.optional is None:
                param.optional = sig_optional

    return cmd


def _signature_validate(cmd, method):
    # type: (Command, MethodType) -> None
    """Ensure that a Command conforms to the method signature

    This will do some validation and will raise a PluginParamError if there are any
    issues.

    It's expected that this will only be called for Parameters where this makes sense
    (aka top-level Parameters). It doesn't make sense to call this for model Parameters,
    so you shouldn't do that.

    Args:
        cmd: Command to validate
        method: Target method object

    Returns:
        None

    Raises:
        PluginParamError: There was a validation problem
    """
    for param in cmd.parameters:
        sig_param = None
        has_kwargs = False

        for p in signature(method).parameters.values():
            if p.name == param.key:
                sig_param = p
            if p.kind == InspectParameter.VAR_KEYWORD:
                has_kwargs = True

        # Couldn't find the parameter. That's OK if this parameter is meant to be part
        # of the **kwargs AND the function has a **kwargs parameter.
        if sig_param is None:
            if not param.is_kwarg:
                raise PluginParamError(
                    "Parameter was not not marked as part of kwargs and wasn't found "
                    "in the method signature (should is_kwarg be True?)"
                )
            elif not has_kwargs:
                raise PluginParamError(
                    "Parameter was declared as a kwarg (is_kwarg=True) but the method "
                    "signature does not declare a **kwargs parameter"
                )

        # Cool, found the parameter. Just verify that it's not pure positional and that
        # it's not marked as part of kwargs.
        else:
            if param.is_kwarg:
                raise PluginParamError(
                    "Parameter was marked as part of kwargs but was found in the "
                    "method signature (should is_kwarg be False?)"
                )

            # I don't think this is even possible in Python < 3.8
            if sig_param.kind == InspectParameter.POSITIONAL_ONLY:
                raise PluginParamError(
                    "Sorry, positional-only type parameters are not supported"
                )


# Alias the old names for compatibility
# This isn't deprecated, see https://github.com/beer-garden/beer-garden/issues/927
system = client


def command_registrar(*args, **kwargs):
    """
    .. deprecated: 3.0
        Will be removed in 4.0. Use ``@system`` instead.
    """
    _deprecate(
        "Looks like you're using the '@command_registrar' decorator. Heads up - this "
        "name will be removed in version 4.0, please use '@system' instead. Thanks!"
    )
    return system(*args, **kwargs)


def register(*args, **kwargs):
    """
    .. deprecated: 3.0
        Will be removed in 4.0. Use ``@command`` instead.
    """
    _deprecate(
        "Looks like you're using the '@register' decorator. Heads up - this name will "
        "be removed in version 4.0, please use '@command' instead. Thanks!"
    )
    return command(*args, **kwargs)


def plugin_param(*args, **kwargs):
    """
    .. deprecated: 3.0
        Will be removed in 4.0. Use ``@parameter`` instead.
    """
    _deprecate(
        "Looks like you're using the '@plugin_param' decorator. Heads up - this name "
        "will be removed in version 4.0, please use '@parameter' instead. Thanks!"
    )
    return parameter(*args, **kwargs)



# -*- coding: utf-8 -*-

import abc
from typing import Any

import six

from brewtils.models import Parameter, Resolvable


@six.add_metaclass(abc.ABCMeta)
class ResolverBase(object):
    """Base for all Resolver implementations"""

    def should_upload(self, value, definition):
        # type: (Any, Parameter) -> bool
        pass

    def upload(self, value, definition):
        # type: (Any, Parameter) -> Resolvable
        pass

    def should_download(self, value, definition):
        # type: (Any, Parameter) -> bool
        pass

    def download(self, value, definition):
        # type: (Resolvable, Parameter) -> Any
        pass



# -*- coding: utf-8 -*-

import calendar
import datetime
from functools import partial

import marshmallow
import simplejson
from marshmallow import Schema, fields, post_load, pre_load
from marshmallow.utils import UTC
from marshmallow_polyfield import PolyField

__all__ = [
    "SystemSchema",
    "InstanceSchema",
    "CommandSchema",
    "ParameterSchema",
    "RequestSchema",
    "RequestFileSchema",
    "FileSchema",
    "FileChunkSchema",
    "FileStatusSchema",
    "PatchSchema",
    "LoggingConfigSchema",
    "EventSchema",
    "QueueSchema",
    "PrincipalSchema",
    "RoleSchema",
    "RefreshTokenSchema",
    "JobSchema",
    "DateTriggerSchema",
    "IntervalTriggerSchema",
    "CronTriggerSchema",
    "FileTriggerSchema",
    "GardenSchema",
    "OperationSchema",
]

# This will be updated after all the schema classes are defined
model_schema_map = {}


def _serialize_model(_, obj, type_field=None, allowed_types=None):
    model_type = getattr(obj, type_field)

    if model_type not in model_schema_map or (
        allowed_types and model_type not in allowed_types
    ):
        raise TypeError("Invalid model type %s" % model_type)

    return model_schema_map.get(model_type)()


def _deserialize_model(_, data, type_field=None, allowed_types=None):
    if data[type_field] not in model_schema_map or (
        allowed_types and data[type_field] not in allowed_types
    ):
        raise TypeError("Invalid payload type %s" % data[type_field])

    return model_schema_map.get(data[type_field])()


class ModelField(PolyField):
    """Field representing a Brewtils model

    Args:
        type_field: Schema field that contains the type information for this field
        allowed_types: A list of allowed model type strings
        **kwargs: Will be passed to the superclass

    """

    def __init__(self, type_field="payload_type", allowed_types=None, **kwargs):
        super(ModelField, self).__init__(
            serialization_schema_selector=partial(
                _serialize_model, type_field=type_field, allowed_types=allowed_types
            ),
            deserialization_schema_selector=partial(
                _deserialize_model, type_field=type_field, allowed_types=allowed_types
            ),
            **kwargs
        )


class DateTime(fields.DateTime):
    """Class that adds methods for (de)serializing DateTime fields as an epoch"""

    def __init__(self, format="epoch", **kwargs):
        self.DATEFORMAT_SERIALIZATION_FUNCS["epoch"] = self.to_epoch
        self.DATEFORMAT_DESERIALIZATION_FUNCS["epoch"] = self.from_epoch
        super(DateTime, self).__init__(format=format, **kwargs)

    @staticmethod
    def to_epoch(dt, localtime=False):
        # If already in epoch form just return it
        if isinstance(dt, int):
            return dt

        if localtime and dt.tzinfo is not None:
            localized = dt
        else:
            if dt.tzinfo is None:
                localized = UTC.localize(dt)
            else:
                localized = dt.astimezone(UTC)
        return (calendar.timegm(localized.timetuple()) * 1000) + int(
            localized.microsecond / 1000
        )

    @staticmethod
    def from_epoch(epoch):
        # If already in datetime form just return it
        if isinstance(epoch, datetime.datetime):
            return epoch

        # utcfromtimestamp will correctly parse milliseconds in Python 3,
        # but in Python 2 we need to help it
        seconds, millis = divmod(epoch, 1000)
        return datetime.datetime.utcfromtimestamp(seconds).replace(
            microsecond=millis * 1000
        )


class BaseSchema(Schema):
    class Meta:
        version_nums = marshmallow.__version__.split(".")
        if int(version_nums[0]) <= 2 and int(version_nums[1]) < 17:
            json_module = simplejson
        else:
            render_module = simplejson

    def __init__(self, strict=True, **kwargs):
        super(BaseSchema, self).__init__(strict=strict, **kwargs)

    @post_load
    def make_object(self, data):
        try:
            model_class = self.context["models"][self.__class__.__name__]
        except KeyError:
            return data

        return model_class(**data)

    @classmethod
    def get_attribute_names(cls):
        return [
            key
            for key, value in cls._declared_fields.items()
            if isinstance(value, fields.FieldABC)
        ]


class ChoicesSchema(BaseSchema):
    type = fields.Str(allow_none=True)
    display = fields.Str(allow_none=True)
    value = fields.Raw(allow_none=True, many=True)
    strict = fields.Bool(allow_none=True, default=False)
    details = fields.Dict(allow_none=True)


class ParameterSchema(BaseSchema):
    key = fields.Str(allow_none=True)
    type = fields.Str(allow_none=True)
    multi = fields.Bool(allow_none=True)
    display_name = fields.Str(allow_none=True)
    optional = fields.Bool(allow_none=True)
    default = fields.Raw(allow_none=True)
    description = fields.Str(allow_none=True)
    choices = fields.Nested("ChoicesSchema", allow_none=True, many=False)
    parameters = fields.Nested("self", many=True, allow_none=True)
    nullable = fields.Bool(allow_none=True)
    maximum = fields.Int(allow_none=True)
    minimum = fields.Int(allow_none=True)
    regex = fields.Str(allow_none=True)
    form_input_type = fields.Str(allow_none=True)
    type_info = fields.Dict(allow_none=True)


class CommandSchema(BaseSchema):
    name = fields.Str(allow_none=True)
    description = fields.Str(allow_none=True)
    parameters = fields.Nested("ParameterSchema", many=True)
    command_type = fields.Str(allow_none=True)
    output_type = fields.Str(allow_none=True)
    schema = fields.Dict(allow_none=True)
    form = fields.Dict(allow_none=True)
    template = fields.Str(allow_none=True)
    icon_name = fields.Str(allow_none=True)
    hidden = fields.Boolean(allow_none=True)
    metadata = fields.Dict(allow_none=True)


class InstanceSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    name = fields.Str(allow_none=True)
    description = fields.Str(allow_none=True)
    status = fields.Str(allow_none=True)
    status_info = fields.Nested("StatusInfoSchema", allow_none=True)
    queue_type = fields.Str(allow_none=True)
    queue_info = fields.Dict(allow_none=True)
    icon_name = fields.Str(allow_none=True)
    metadata = fields.Dict(allow_none=True)


class SystemSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    name = fields.Str(allow_none=True)
    description = fields.Str(allow_none=True)
    version = fields.Str(allow_none=True)
    max_instances = fields.Integer(allow_none=True)
    icon_name = fields.Str(allow_none=True)
    instances = fields.Nested("InstanceSchema", many=True, allow_none=True)
    commands = fields.Nested("CommandSchema", many=True, allow_none=True)
    display_name = fields.Str(allow_none=True)
    metadata = fields.Dict(allow_none=True)
    namespace = fields.Str(allow_none=True)
    local = fields.Bool(allow_none=True)
    template = fields.Str(allow_none=True)


class RequestFileSchema(BaseSchema):
    storage_type = fields.Str(allow_none=True)
    filename = fields.Str(allow_none=True)
    id = fields.Str(allow_none=False)


class FileSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    owner_id = fields.Str(allow_none=True)
    owner_type = fields.Str(allow_none=True)
    owner = fields.Raw(allow_none=True)
    job = fields.Nested("JobSchema", allow_none=True)
    request = fields.Nested("RequestSchema", allow_none=True)
    updated_at = DateTime(allow_none=True, format="epoch", example="1500065932000")
    file_name = fields.Str(allow_none=True)
    file_size = fields.Int(allow_none=False)
    chunks = fields.Dict(allow_none=True)
    chunk_size = fields.Int(allow_none=False)


class FileChunkSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    file_id = fields.Str(allow_none=False)
    offset = fields.Int(allow_none=False)
    data = fields.Str(allow_none=False)
    owner = fields.Nested("FileSchema", allow_none=True)


class FileStatusSchema(BaseSchema):
    # Top-level file info
    file_id = fields.Str(allow_none=True)
    updated_at = fields.Str(allow_none=True)
    file_name = fields.Str(allow_none=True)
    file_size = fields.Int(allow_none=True)
    chunk_size = fields.Int(allow_none=True)
    chunks = fields.Dict(allow_none=True)
    owner_id = fields.Str(allow_none=True)
    owner_type = fields.Str(allow_none=True)
    # Chunk info
    chunk_id = fields.Str(allow_none=True)
    offset = fields.Int(allow_none=True)
    data = fields.Str(allow_none=True)
    # Validation metadata
    valid = fields.Bool(allow_none=True)
    missing_chunks = fields.List(fields.Int(), allow_none=True)
    expected_number_of_chunks = fields.Int(allow_none=True)
    expected_max_size = fields.Int(allow_none=True)
    number_of_chunks = fields.Int(allow_none=True)
    size_ok = fields.Bool(allow_none=True)
    chunks_ok = fields.Bool(allow_none=True)
    operation_complete = fields.Bool(allow_none=True)
    message = fields.Str(allow_none=True)


class RequestTemplateSchema(BaseSchema):
    """Used as a base class for request and a request template for jobs."""

    system = fields.Str(allow_none=True)
    system_version = fields.Str(allow_none=True)
    instance_name = fields.Str(allow_none=True)
    namespace = fields.Str(allow_none=True)
    command = fields.Str(allow_none=True)
    command_type = fields.Str(allow_none=True)
    parameters = fields.Dict(allow_none=True)
    comment = fields.Str(allow_none=True)
    metadata = fields.Dict(allow_none=True)
    output_type = fields.Str(allow_none=True)


class RequestSchema(RequestTemplateSchema):
    id = fields.Str(allow_none=True)
    parent = fields.Nested("self", exclude=("children",), allow_none=True)
    children = fields.Nested(
        "self", exclude=("parent", "children"), many=True, default=None, allow_none=True
    )
    output = fields.Str(allow_none=True)
    hidden = fields.Boolean(allow_none=True)
    status = fields.Str(allow_none=True)
    error_class = fields.Str(allow_none=True)
    created_at = DateTime(allow_none=True, format="epoch", example="1500065932000")
    updated_at = DateTime(allow_none=True, format="epoch", example="1500065932000")
    has_parent = fields.Bool(allow_none=True)
    requester = fields.String(allow_none=True)


class StatusInfoSchema(BaseSchema):
    heartbeat = DateTime(allow_none=True, format="epoch", example="1500065932000")


class PatchSchema(BaseSchema):
    operation = fields.Str(allow_none=True)
    path = fields.Str(allow_none=True)
    value = fields.Raw(allow_none=True)

    @pre_load(pass_many=True)
    def unwrap_envelope(self, data, many):
        """Helper function for parsing the different patch formats.

        This exists because previously multiple patches serialized like::

            {
                "operations": [
                    {"operation": "replace", ...},
                    {"operation": "replace", ...}
                    ...
                ]
            }

        But we also wanted to be able to handle a simple list::

            [
                {"operation": "replace", ...},
                {"operation": "replace", ...}
                ...
            ]

        Patches are now (as of v3) serialized as the latter. Prior to v3 they were
        serialized as the former.
        """
        if isinstance(data, list):
            return data
        elif "operations" in data:
            return data["operations"]
        else:
            return [data]


class LoggingConfigSchema(BaseSchema):
    level = fields.Str(allow_none=True)
    formatters = fields.Dict(allow_none=True)
    handlers = fields.Dict(allow_none=True)


class EventSchema(BaseSchema):

    name = fields.Str(allow_none=True)
    namespace = fields.Str(allow_none=True)
    garden = fields.Str(allow_none=True)
    metadata = fields.Dict(allow_none=True)
    timestamp = DateTime(allow_none=True, format="epoch", example="1500065932000")

    payload_type = fields.Str(allow_none=True)
    payload = ModelField(allow_none=True)

    error = fields.Bool(allow_none=True)
    error_message = fields.Str(allow_none=True)


class QueueSchema(BaseSchema):
    name = fields.Str(allow_none=True)
    system = fields.Str(allow_none=True)
    version = fields.Str(allow_none=True)
    instance = fields.Str(allow_none=True)
    system_id = fields.Str(allow_none=True)
    display = fields.Str(allow_none=True)
    size = fields.Integer(allow_none=True)


class PrincipalSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    username = fields.Str(allow_none=True)
    roles = fields.Nested("RoleSchema", many=True, allow_none=True)
    permissions = fields.List(fields.Str(), allow_none=True)
    preferences = fields.Dict(allow_none=True)
    metadata = fields.Dict(allow_none=True)


class RoleSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    name = fields.Str(allow_none=True)
    description = fields.Str(allow_none=True)
    roles = fields.Nested("self", many=True, allow_none=True)
    permissions = fields.List(fields.Str(), allow_none=True)


class RefreshTokenSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    issued = DateTime(allow_none=True, format="epoch", example="1500065932000")
    expires = DateTime(allow_none=True, format="epoch", example="1500065932000")
    payload = fields.Dict(allow_none=True)


class DateTriggerSchema(BaseSchema):
    run_date = DateTime(allow_none=True, format="epoch", example="1500065932000")
    timezone = fields.Str(allow_none=True)


class IntervalTriggerSchema(BaseSchema):
    weeks = fields.Int(allow_none=True)
    days = fields.Int(allow_none=True)
    hours = fields.Int(allow_none=True)
    minutes = fields.Int(allow_none=True)
    seconds = fields.Int(allow_none=True)
    start_date = DateTime(allow_none=True, format="epoch", example="1500065932000")
    end_date = DateTime(allow_none=True, format="epoch", example="1500065932000")
    timezone = fields.Str(allow_none=True)
    jitter = fields.Int(allow_none=True)
    reschedule_on_finish = fields.Bool(allow_none=True)


class CronTriggerSchema(BaseSchema):
    year = fields.Str(allow_none=True)
    month = fields.Str(allow_none=True)
    day = fields.Str(allow_none=True)
    week = fields.Str(allow_none=True)
    day_of_week = fields.Str(allow_none=True)
    hour = fields.Str(allow_none=True)
    minute = fields.Str(allow_none=True)
    second = fields.Str(allow_none=True)
    start_date = DateTime(allow_none=True, format="epoch", example="1500065932000")
    end_date = DateTime(allow_none=True, format="epoch", example="1500065932000")
    timezone = fields.Str(allow_none=True)
    jitter = fields.Int(allow_none=True)


class FileTriggerSchema(BaseSchema):
    pattern = fields.List(fields.Str(), allow_none=True)
    path = fields.Str(allow_none=True)
    recursive = fields.Bool(allow_none=True)
    callbacks = fields.Dict(fields.Bool(), allow_none=True)


class GardenSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    name = fields.Str(allow_none=True)
    status = fields.Str(allow_none=True)
    status_info = fields.Nested("StatusInfoSchema", allow_none=True)
    connection_type = fields.Str(allow_none=True)
    connection_params = fields.Dict(allow_none=True)
    namespaces = fields.List(fields.Str(), allow_none=True)
    systems = fields.Nested("SystemSchema", many=True, allow_none=True)


class JobSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    name = fields.Str(allow_none=True)
    trigger_type = fields.Str(allow_none=True)
    trigger = ModelField(
        type_field="trigger_type",
        allowed_types=["interval", "date", "cron", "file"],
        allow_none=True,
    )
    request_template = fields.Nested("RequestTemplateSchema", allow_none=True)
    misfire_grace_time = fields.Int(allow_none=True)
    coalesce = fields.Bool(allow_none=True)
    next_run_time = DateTime(allow_none=True, format="epoch", example="1500065932000")
    success_count = fields.Int(allow_none=True)
    error_count = fields.Int(allow_none=True)
    status = fields.Str(allow_none=True)
    max_instances = fields.Int(allow_none=True)
    timeout = fields.Int(allow_none=True)


class OperationSchema(BaseSchema):
    model_type = fields.Str(allow_none=True)
    model = ModelField(allow_none=True, type_field="model_type")

    args = fields.List(fields.Str(), allow_none=True)
    kwargs = fields.Dict(allow_none=True)

    target_garden_name = fields.Str(allow_none=True)
    source_garden_name = fields.Str(allow_none=True)
    operation_type = fields.Str(allow_none=True)


class RunnerSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    name = fields.Str(allow_none=True)
    path = fields.Str(allow_none=True)
    instance_id = fields.Str(allow_none=True)
    stopped = fields.Boolean(allow_none=True)
    dead = fields.Boolean(allow_none=True)
    restart = fields.Boolean(allow_none=True)


class ResolvableSchema(BaseSchema):
    id = fields.Str(allow_none=True)
    type = fields.Str(allow_none=True)
    storage = fields.Str(allow_none=True)
    details = fields.Dict(allow_none=True)


model_schema_map.update(
    {
        "Choices": ChoicesSchema,
        "Command": CommandSchema,
        "CronTrigger": CronTriggerSchema,
        "DateTrigger": DateTriggerSchema,
        "Event": EventSchema,
        "FileTrigger": FileTriggerSchema,
        "Garden": GardenSchema,
        "Instance": InstanceSchema,
        "IntervalTrigger": IntervalTriggerSchema,
        "Job": JobSchema,
        "LoggingConfig": LoggingConfigSchema,
        "Queue": QueueSchema,
        "Parameter": ParameterSchema,
        "PatchOperation": PatchSchema,
        "Principal": PrincipalSchema,
        "RefreshToken": RefreshTokenSchema,
        "Request": RequestSchema,
        "RequestFile": RequestFileSchema,
        "File": FileSchema,
        "FileChunk": FileChunkSchema,
        "FileStatus": FileStatusSchema,
        "RequestTemplate": RequestTemplateSchema,
        "Role": RoleSchema,
        "System": SystemSchema,
        "Operation": OperationSchema,
        "Runner": RunnerSchema,
        "Resolvable": ResolvableSchema,
        # Compatibility for the Job trigger types
        "interval": IntervalTriggerSchema,
        "date": DateTriggerSchema,
        "cron": CronTriggerSchema,
        "file": FileTriggerSchema,
    }
)
